"""gjson custom exceptions module."""
from typing import Any


class GJSONError(Exception):
    """Raised by the gjson module on error while performing queries or converting to JSON."""


class GJSONParseError(GJSONError):
    """Raised when there is an error parsing the query string, with nicer representation of the error."""

    def __init__(self, *args: Any, query: str, position: int):
        """Initialize the exception with the additional data of the query part.

        Arguments:
            *args: all positional arguments like any regular exception.
            query: the full query that generated the parse error.
            position: the position in the query string where the parse error occurred.

        """
        super().__init__(*args)
        self.query = query
        self.position = position

    def __str__(self) -> str:
        """Return a custom representation of the error.

        Returns:
            the whole query string with a clear indication on where the error occurred.

        """
        default = super().__str__()
        line = '-' * (self.position + 7)  # 7 is for the lenght of 'Query: '
        return f'{default}\nQuery: {self.query}\n{line}^'


class GJSONInvalidSyntaxError(GJSONParseError):
    """Raised when there is a query with an invalid syntax."""



"""CLI test module."""
import argparse
import io
import json

import pytest

from gjson._cli import cli
from gjson.exceptions import GJSONParseError

from .test_init import INPUT_JSON, INPUT_LINES, INPUT_LINES_WITH_ERRORS


def test_cli_stdin(monkeypatch, capsys):
    """It should read the data from stdin and query it."""
    monkeypatch.setattr('sys.stdin', io.StringIO(INPUT_JSON))
    ret = cli(['-', 'name.first'])
    assert ret == 0
    captured = capsys.readouterr()
    assert captured.out == '"Tom"\n'
    assert not captured.err


def test_cli_file(tmp_path, capsys):
    """It should read the data from the provided file and query it."""
    data_file = tmp_path / 'input.json'
    data_file.write_text(INPUT_JSON)
    ret = cli([str(data_file), 'name.first'])
    assert ret == 0
    captured = capsys.readouterr()
    assert captured.out == '"Tom"\n'
    assert not captured.err


@pytest.mark.parametrize('query', (
    'a\tkey',
    '..0.a\tkey',
))
def test_cli_file_with_control_chars(query, tmp_path, capsys):
    """It should read the data from the provided file and query it."""
    data_file = tmp_path / 'input.json'
    data_file.write_text('{"a\tkey": "a\tvalue"}')
    ret = cli(['-vvv', str(data_file), query])
    assert ret == 0
    captured = capsys.readouterr()
    assert captured.out == '"a\\tvalue"\n'
    assert not captured.err


@pytest.mark.parametrize('query', (
    'a\tkey',
    '..0.a\tkey',
))
def test_cli_stdin_with_control_chars(query, monkeypatch, capsys):
    """It should read the data from stdin and query it."""
    monkeypatch.setattr('sys.stdin', io.StringIO('{"a\tkey": "a\tvalue"}'))
    ret = cli(['-vvv', '-', query])
    assert ret == 0
    captured = capsys.readouterr()
    assert captured.out == '"a\\tvalue"\n'
    assert not captured.err


def test_cli_nonexistent_file(tmp_path, capsys):
    """It should exit with a failure exit code and no output."""
    ret = cli([str(tmp_path / 'nonexistent.json'), 'name.first'])
    assert ret == 1
    captured = capsys.readouterr()
    assert not captured.out
    assert not captured.err


def test_cli_nonexistent_file_verbosity_1(tmp_path, capsys):
    """It should exit with a failure exit code and print the error message."""
    ret = cli(['-v', str(tmp_path / 'nonexistent.json'), 'name.first'])
    assert ret == 1
    captured = capsys.readouterr()
    assert not captured.out
    assert captured.err.startswith("ArgumentTypeError: can't open")
    assert 'nonexistent.json' in captured.err


def test_cli_nonexistent_file_verbosity_2(tmp_path):
    """It should raise the exception and print the full traceback."""
    with pytest.raises(
            argparse.ArgumentTypeError, match=r"can't open .*/nonexistent.json.* No such file or directory"):
        cli(['-vv', str(tmp_path / 'nonexistent.json'), 'name.first'])


def test_cli_stdin_query_verbosity_1(monkeypatch, capsys):
    """It should exit with a failure exit code and print the error message."""
    monkeypatch.setattr('sys.stdin', io.StringIO(INPUT_JSON))
    ret = cli(['-v', '-', 'nonexistent'])
    assert ret == 1
    captured = capsys.readouterr()
    assert not captured.out
    assert captured.err == ('GJSONParseError: Mapping object does not have key `nonexistent`.\n'
                            'Query: nonexistent\n-------^\n')


def test_cli_stdin_query_verbosity_2(monkeypatch):
    """It should exit with a failure exit code and print the full traceback."""
    monkeypatch.setattr('sys.stdin', io.StringIO(INPUT_JSON))
    with pytest.raises(GJSONParseError, match=r'Mapping object does not have key `nonexistent`.'):
        cli(['-vv', '-', 'nonexistent'])


def test_cli_lines_ok(monkeypatch, capsys):
    """It should apply the same query to each line."""
    monkeypatch.setattr('sys.stdin', io.StringIO(INPUT_LINES))
    ret = cli(['--lines', '-', 'name'])
    assert ret == 0
    captured = capsys.readouterr()
    assert captured.out == '"Gilbert"\n"Alexa"\n"May"\n"Deloise"\n'
    assert not captured.err


def test_cli_lines_failed_lines_verbosity_0(monkeypatch, capsys):
    """It should keep going with the other lines and just skip the failed line."""
    monkeypatch.setattr('sys.stdin', io.StringIO(INPUT_LINES_WITH_ERRORS))
    ret = cli(['--lines', '-', 'name'])
    assert ret == 1
    captured = capsys.readouterr()
    assert captured.out == '"Gilbert"\n"Deloise"\n'
    assert not captured.err


def test_cli_lines_failed_lines_verbosity_1(monkeypatch, capsys):
    """It should keep going with the other lines printing an error for the failed lines."""
    monkeypatch.setattr('sys.stdin', io.StringIO(INPUT_LINES_WITH_ERRORS))
    ret = cli(['-v', '--lines', '-', 'name'])
    assert ret == 1
    captured = capsys.readouterr()
    assert captured.out == '"Gilbert"\n"Deloise"\n'
    assert captured.err.count('JSONDecodeError') == 2


def test_cli_lines_failed_lines_verbosity_2(monkeypatch):
    """It should interrupt the processing and print the full traceback."""
    monkeypatch.setattr('sys.stdin', io.StringIO(INPUT_LINES_WITH_ERRORS))
    with pytest.raises(
            json.decoder.JSONDecodeError, match=r'Expecting property name enclosed in double quotes'):
        cli(['-vv', '--lines', '-', 'name'])


def test_cli_lines_double_dot_query(monkeypatch, capsys):
    """It should encapsulate each line in an array to allow queries."""
    monkeypatch.setattr('sys.stdin', io.StringIO(INPUT_LINES))
    ret = cli(['--lines', '..#(age>45).name'])
    assert ret == 1
    captured = capsys.readouterr()
    assert captured.out == '"Gilbert"\n"May"\n'
    assert not captured.err


def test_cli_double_dot_query_ok(monkeypatch, capsys):
    """It should encapsulate the input in an array and apply the query to the array."""
    monkeypatch.setattr('sys.stdin', io.StringIO(INPUT_LINES))
    ret = cli(['-', '..#.name'])
    assert ret == 0
    captured = capsys.readouterr()
    assert captured.out == '["Gilbert", "Alexa", "May", "Deloise"]\n'
    assert not captured.err


def test_cli_double_dot_query_failed_lines_verbosity_0(monkeypatch, capsys):
    """It should encapsulate the input in an array skipping failing lines."""
    monkeypatch.setattr('sys.stdin', io.StringIO(INPUT_LINES_WITH_ERRORS))
    ret = cli(['-', '..#.name'])
    assert ret == 1
    captured = capsys.readouterr()
    assert not captured.out
    assert not captured.err


def test_cli_double_dot_query_failed_lines_verbosity_1(monkeypatch, capsys):
    """It should encapsulate the input in an array skipping failing lines and printing an error for each failure."""
    monkeypatch.setattr('sys.stdin', io.StringIO(INPUT_LINES_WITH_ERRORS))
    ret = cli(['-v', '-', '..#.name'])
    assert ret == 1
    captured = capsys.readouterr()
    assert not captured.out
    assert captured.err.startswith('JSONDecodeError: Expecting property name enclosed in double quotes')


def test_cli_double_dot_query_failed_lines_verbosity_2(monkeypatch):
    """It should interrupt the execution at the first invalid line and exit printing the traceback."""
    monkeypatch.setattr('sys.stdin', io.StringIO(INPUT_LINES_WITH_ERRORS))
    with pytest.raises(json.decoder.JSONDecodeError, match=r'Expecting property name enclosed in double quotes'):
        cli(['-vv', '-', '..#.name'])



"""Typing protocols used by the gjson package."""
from typing import Any, Protocol, runtime_checkable


@runtime_checkable
class ModifierProtocol(Protocol):
    """Callback protocol for the custom modifiers."""

    def __call__(self, options: dict[str, Any], obj: Any, *, last: bool) -> Any:
        """To register a custom modifier a callable that adhere to this protocol must be provided.

        Examples:
            Register a custom modifier that sums all the numbers in a list:

                >>> import gjson
                >>> data = [1, 2, 3, 4, 5]
                >>> def custom_sum(options, obj, *, last):
                ...     # insert sanity checks code here
                ...     return sum(obj)
                ...
                >>> gjson_obj = gjson.GJSON(data)
                >>> gjson_obj.register_modifier('sum', custom_sum)
                >>> gjson_obj.get('@sum')
                15

        Arguments:
            options: a dictionary of options. If no options are present in the query the callable will be called with
                an empty dictionary as options. The modifier can supports any number of options, or none.
            obj: the current object already modifier by any previous parts of the query.
            last: :py:data:`True` if the modifier is the last element in the query or :py:data:`False` otherwise.

        Raises:
            Exception: any exception that might be raised by the callable is catched by gjson and re-raised as a
                :py:class:`gjson.GJSONError` exception to ensure that the normal gjson behaviour is respected according
                to the selected verbosity (CLI) or ``quiet`` parameter (Python library).

        Returns:
            the resulting object after applying the modifier.

        """



"""GJSON module."""
import argparse
import json
import sys
from collections.abc import Sequence
from typing import Any, IO, Optional

from gjson import get, GJSONError


def cli(argv: Optional[Sequence[str]] = None) -> int:  # noqa: MC0001
    """Command line entry point to run gjson as a CLI tool.

    Arguments:
        argv: a sequence of CLI arguments to parse. If not set they will be read from sys.argv.

    Returns:
        The CLI exit code to use.

    Raises:
        OSError: for system-related error, including I/O failures.
        json.JSONDecodeError: when the input data is not a valid JSON.
        gjson.GJSONError: for any query-related error in gjson.

    """
    parser = get_parser()
    args = parser.parse_args(argv)

    encapsulate = False
    if args.query.startswith('..'):
        args.query = args.query[2:]
        encapsulate = True

    # Use argparse.FileType here instead of putting it as type in the --file argument parsing, to allow to handle the
    # verbosity in case of error and make sure the file is always closed in case other arguments fail the validation.
    try:
        args.file = argparse.FileType(encoding='utf-8', errors='surrogateescape')(args.file)
    except (OSError, argparse.ArgumentTypeError) as ex:
        if args.verbose == 1:
            print(f'{ex.__class__.__name__}: {ex}', file=sys.stderr)
        elif args.verbose >= 2:
            raise

        return 1

    # Reconfigure __stdin__ and __stdout__ instead of stdin and stdout because the latters are TextIO and could not
    # have the reconfigure() method if re-assigned, while reconfigure() is part of TextIOWrapper.
    # See also: https://github.com/python/typeshed/pull/8171
    sys.__stdin__.reconfigure(errors='surrogateescape')
    sys.__stdout__.reconfigure(errors='surrogateescape')

    def _execute(line: str, file_obj: Optional[IO[Any]]) -> int:
        try:
            if encapsulate:
                if line:
                    input_data = [json.loads(line, strict=False)]
                elif file_obj is not None:
                    input_data = []
                    for input_line in file_obj:
                        if input_line.strip():
                            input_data.append(json.loads(input_line, strict=False))
            else:
                if line:
                    input_data = json.loads(line, strict=False)
                elif file_obj is not None:
                    input_data = json.load(file_obj, strict=False)

            result = get(input_data, args.query, as_str=True)
            exit_code = 0
        except (json.JSONDecodeError, GJSONError) as ex:
            result = ''
            exit_code = 1
            if args.verbose == 1:
                print(f'{ex.__class__.__name__}: {ex}', file=sys.stderr)
            elif args.verbose >= 2:
                raise

        if result:
            print(result)

        return exit_code

    if args.lines:
        exit_code = 0
        for line in args.file:
            line = line.strip()
            if not line:
                continue
            ret = _execute(line, None)
            if ret > exit_code:
                exit_code = ret
    else:
        exit_code = _execute('', args.file)

    return exit_code


def get_parser() -> argparse.ArgumentParser:
    """Get the CLI argument parser.

    Returns:
        the argument parser for the CLI.

    """
    parser = argparse.ArgumentParser(
        prog='gjson',
        description=('A simple way to filter and extract data from JSON-like data structures. Python porting of the '
                     'Go GJSON package.'),
        epilog='See also the full documentation available at https://volans-.github.io/gjson-py/index.html',
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    parser.add_argument('-v', '--verbose', action='count', default=0,
                        help=('Verbosity level. By default on error no output will be printed. Use -v to get the '
                              'error message to stderr and -vv to get the full traceback.'))
    parser.add_argument('-l', '--lines', action='store_true',
                        help='Treat the input as JSON Lines, parse each line and apply the query to each line.')
    # argparse.FileType is used later to parse this argument.
    parser.add_argument('file', default='-', nargs='?',
                        help='Input JSON file to query. Reads from stdin if the argument is missing or set to "-".')
    parser.add_argument('query', help='A GJSON query to apply to the input data.')

    return parser



"""GJSON test module."""
# pylint: disable=attribute-defined-outside-init
import json
import re

from collections.abc import Mapping
from math import isnan

import pytest

import gjson
from gjson._gjson import MODIFIER_NAME_RESERVED_CHARS


INPUT_JSON = """
{
    "name": {"first": "Tom", "last": "Anderson"},
    "age":37,
    "children": ["Sara","Alex","Jack"],
    "fav.movie": "Deer Hunter",
    "friends": [
        {"first": "Dale", "last": "Murphy", "age": 44, "nets": ["ig", "fb", "tw"]},
        {"first": "Roger", "last": "Craig", "age": 68, "nets": ["fb", "tw"]},
        {"first": "Jane", "last": "Murphy", "age": 47, "nets": ["ig", "tw"]}
    ]
}
"""
INPUT_OBJECT = json.loads(INPUT_JSON)
INPUT_LIST = json.loads("""
[
    {"first": "Dale"},
    {"first": "Jane"},
    {"last": "Murphy"}
]
""")
INPUT_ESCAPE = json.loads("""
{
    "test": {
        "*":"valZ",
        "*v":"val0",
        "keyv*":"val1",
        "key*v":"val2",
        "keyv?":"val3",
        "key?v":"val4",
        "keyv.":"val5",
        "key.v":"val6",
        "keyk*":{"key?":"val7"},
        "1key":"val8"
    }
}
""")
# This json block is poorly formed on purpose.
INPUT_BASIC = json.loads("""
  {"age":100, "name2":{"here":"B\\\\\\"R"},
    "noop":{"what is a wren?":"a bird"},
    "happy":true,"immortal":false,
    "items":[1,2,3,{"tags":[1,2,3],"points":[[1,2],[3,4]]},4,5,6,7],
    "arr":["1",2,"3",{"hello":"world"},"4",5],
    "vals":[1,2,3],"name":{"first":"tom","last":null},
    "created":"2014-05-16T08:28:06.989Z",
    "loggy":{
        "programmers": [
            {
                "firstName": "Brett",
                "lastName": "McLaughlin",
                "email": "aaaa",
                "tag": "good"
            },
            {
                "firstName": "Jason",
                "lastName": "Hunter",
                "email": "bbbb",
                "tag": "bad"
            },
            {
                "firstName": "Elliotte",
                "lastName": "Harold",
                "email": "cccc",
                "tag": "good"
            },
            {
                "firstName": 1002.3,
                "age": 101
            }
        ]
    },
    "lastly":{"end...ing":"soon","yay":"final"}
}
""")
INPUT_LINES = """
{"name": "Gilbert", "age": 61}
{"name": "Alexa", "age": 34}
{"name": "May", "age": 57}
{"name": "Deloise", "age": 44}
"""
INPUT_LINES_WITH_ERRORS = """
{"name": "Gilbert", "age": 61}
{invalid
{invalid
{"name": "Deloise", "age": 44}
"""
INPUT_TRUTHINESS = json.loads("""
{
  "vals": [
    { "a": 1, "b": true },
    { "a": 2, "b": true },
    { "a": 3, "b": false },
    { "a": 4, "b": "0" },
    { "a": 5, "b": 0 },
    { "a": 6, "b": "1" },
    { "a": 7, "b": 1 },
    { "a": 8, "b": "true" },
    { "a": 9, "b": false },
    { "a": 10, "b": null },
    { "a": 11 }
  ]
}
""")
INPUT_SUM_N = json.loads("""
[
    {"key": "a", "value": 1, "other": "value"},
    {"key": "b", "value": 2},
    {"key": "c", "value": 3, "other": "value"},
    {"key": "a", "value": 7},
    {"key": "b", "value": 1.5},
    {"key": "d", "value": 4},
    {"key": "c", "value": 9}
]
""")
INPUT_NESTED_QUERIES = json.loads("""
{
    "key": [
        {"level1": [{"level2": [{"level3": [1, 2]}]}]},
        {"level1": [{"level2": [{"level3": [2, 3]}]}]},
        [[{"level3": [1, 2]}], [{"level3": [2, 3]}]],
        [[{"level3": [2, 3]}], [{"level3": [3, 4]}]],
        {"another": [{"level2": [{"level3": [2, 3]}]}]},
        {"level1": [{"another": [{"level3": [2, 3]}]}]},
        {"level1": [{"level2": [{"another": [2, 3]}]}]},
        [[{"another": [2, 3]}], [{"another": [3, 4]}]],
        "spurious",
        12.34,
        {"mixed": [[{"level4": [1, 2]}]]}
    ]
}
""")


def compare_values(result, expected):
    """Compare results with the expected values ensuring same-order of keys for dictionaries."""
    if isinstance(expected, float):
        if isnan(expected):
            assert isnan(result)
            return

    assert result == expected
    if isinstance(expected, Mapping):
        assert list(result.keys()) == list(expected.keys())


class TestObject:
    """Testing gjson with a basic input object."""

    def setup_method(self):
        """Initialize the test instance."""
        def upper(options, obj, *, last):
            """Custom modifier to return a string upper case."""
            del options
            del last
            if isinstance(obj, list):
                return [i.upper() for i in obj]

            return obj.upper()

        self.object = gjson.GJSON(INPUT_OBJECT)
        self.object.register_modifier('upper', upper)

    @pytest.mark.parametrize('query, expected', (
        # Basic
        ('name.last', 'Anderson'),
        ('name.first', 'Tom'),
        ('age', 37),
        ('children', ['Sara', 'Alex', 'Jack']),
        ('children.0', 'Sara'),
        ('children.1', 'Alex'),
        ('friends.1', {'first': 'Roger', 'last': 'Craig', 'age': 68, 'nets': ['fb', 'tw']}),
        ('friends.1.first', 'Roger'),
        # Wildcards
        ('*.first', 'Tom'),
        ('?a??.first', 'Tom'),
        ('child*.2', 'Jack'),
        ('c?ildren.0', 'Sara'),
        # Escape characters
        (r'fav\.movie', 'Deer Hunter'),
        # Arrays
        ('friends.#', 3),
        ('friends.#.age', [44, 68, 47]),
        ('friends.#.first', ['Dale', 'Roger', 'Jane']),
        # Queries
        ('friends.#(last=="Murphy").first', 'Dale'),
        ('friends.#(last=="Murphy")#.first', ['Dale', 'Jane']),
        ('friends.#(=="Murphy")#', []),
        ('friends.#(=="Mu)(phy")#', []),
        ('friends.#(=="Mur\tphy")#', []),
        ('friends.#(age\\===44)#', []),
        ('friends.#(age>47)#.last', ['Craig']),
        ('friends.#(age>=47)#.last', ['Craig', 'Murphy']),
        ('friends.#(age<47)#.last', ['Murphy']),
        ('friends.#(age<=47)#.last', ['Murphy', 'Murphy']),
        ('friends.#(age==44)#.last', ['Murphy']),
        ('friends.#(age!=44)#.last', ['Craig', 'Murphy']),
        ('friends.#(first%"D*").last', 'Murphy'),
        ('friends.#(first!%"D*").last', 'Craig'),
        ('friends.#(first!%"D???").last', 'Craig'),
        ('friends.#(%0)#', []),
        ('friends.#(>40)#', []),
        ('children.#(!%"*a*")', 'Alex'),
        ('children.#(%"*a*")#', ['Sara', 'Jack']),
        # Nested queries (TODO)
        # ('friends.#(nets.#(=="fb"))#.first', ['Dale', 'Roger']),
        # Tilde in queries (TODO)
        # ('vals.#(b==~true)#.a')
        # Modifiers
        ('children.@reverse', ['Jack', 'Alex', 'Sara']),
        ('children.@reverse.0', 'Jack'),
        ('name.@reverse', {'last': 'Anderson', 'first': 'Tom'}),
        ('age.@reverse', 37),
        ('@keys', ['name', 'age', 'children', 'fav.movie', 'friends']),
        ('name.@values', ['Tom', 'Anderson']),
        ('age.@flatten', 37),
        ('@pretty:{"indent": 4}', INPUT_OBJECT),
        (r'fav\.movie.@pretty:{"indent": 4}', 'Deer Hunter'),
        ('name.@tostr', '{"first": "Tom", "last": "Anderson"}'),
        ('name.@join', {'first': 'Tom', 'last': 'Anderson'}),
        ('age.@join', 37),
        ('children.@join', {}),
        ('children.0.@join', 'Sara'),
        ('friends.@join', {'first': 'Jane', 'last': 'Murphy', 'age': 47, 'nets': ['ig', 'tw']}),
        # Dot vs Pipe
        ('friends.0.first', 'Dale'),
        ('friends|0.first', 'Dale'),
        ('friends.0|first', 'Dale'),
        ('friends|0|first', 'Dale'),
        ('friends|#', 3),
        ('friends.#', 3),
        ('friends.#(last="Murphy")#',
         [{'first': 'Dale', 'last': 'Murphy', 'age': 44, 'nets': ['ig', 'fb', 'tw']},
          {'first': 'Jane', 'last': 'Murphy', 'age': 47, 'nets': ['ig', 'tw']}]),
        ('friends.#(last="Murphy")#.first', ['Dale', 'Jane']),
        ('friends.#(last="Murphy")#.0', []),
        ('friends.#(last="Murphy")#|0', {'first': 'Dale', 'last': 'Murphy', 'age': 44, 'nets': ['ig', 'fb', 'tw']}),
        ('friends.#(last="Murphy")#.#', []),
        ('friends.#(last="Murphy")#|#', 2),
        # Multipaths objects
        ('{}', {}),
        ('{.}', {}),
        ('{.invalid}', {}),
        ('{.invalid,}', {}),
        ('{age}', {'age': 37}),
        (r'{a\ge}', {r'a\ge': 37}),
        (r'{"a\\ge":age}', {r'a\ge': 37}),
        ('{"a\tb":age}', {'a\tb': 37}),
        ('{"key":age}', {'key': 37}),
        ('{age,age}', {'age': 37}),
        ('{age,"years":age}', {'age': 37, 'years': 37}),
        ('{"years":age,age}', {'years': 37, 'age': 37}),
        ('{age,name.first}', {'age': 37, 'first': 'Tom'}),
        ('{invalid,invalid.invalid,age}', {'age': 37}),
        ('{name.first,age,name.last}', {'first': 'Tom', 'age': 37, 'last': 'Anderson'}),
        ('{{age}}', {'_': {'age': 37}}),
        ('{{age},age}', {'_': {'age': 37}, 'age': 37}),
        ('friends.0.{age,nets.#(="ig")}', {'age': 44, "_": 'ig'}),
        ('friends.0.{age,nets.#(="ig"),invalid}', {'age': 44, "_": 'ig'}),
        ('friends.0.{age,nets.#(="ig")#}', {'age': 44, "_": ['ig']}),
        ('friends.#.{age,"key":first}',
         [{'age': 44, 'key': 'Dale'}, {'age': 68, 'key': 'Roger'}, {'age': 47, 'key': 'Jane'}]),
        ('friends.#(age>44)#.{age,"key":first}', [{'age': 68, 'key': 'Roger'}, {'age': 47, 'key': 'Jane'}]),
        ('friends.#(age>44)#.{age,"key":first,invalid}', [{'age': 68, 'key': 'Roger'}, {'age': 47, 'key': 'Jane'}]),
        (r'{age,name.first,fav\.movie}', {'age': 37, 'first': 'Tom', r'fav\.movie': 'Deer Hunter'}),
        ('{age,name.{"name":first,"surname":last},children.@sort}',
         {'age': 37, '_': {'name': 'Tom', 'surname': 'Anderson'}, '@sort': ['Alex', 'Jack', 'Sara']}),
        ('friends.{0.first,1.last,2.age}.@values', ['Dale', 'Craig', 47]),
        ('{friends.{"a":0.{nets.{0}}}}', {'_': {'a': {'_': {'0': 'ig'}}}}),
        ('{friends.{"a":0.{nets.{0,1}}}}', {'_': {'a': {'_': {'0': 'ig', '1': 'fb'}}}}),
        ('friends.#.{age,first|@upper}',
         [{"age": 44, "@upper": "DALE"}, {"age": 68, "@upper": "ROGER"}, {"age": 47, "@upper": "JANE"}]),
        ('{friends.#.{age,"first":first|@upper}|0.first}', {"first": "DALE"}),
        ('{"children":children|@upper,"name":name.first,"age":age}',
         {"children": ["SARA", "ALEX", "JACK"], "name": "Tom", "age": 37}),
        ('friends.#.{age,"first":first.invalid}', [{'age': 44}, {'age': 68}, {'age': 47}]),
        # Multipaths arrays
        ('[]', []),
        ('[.]', []),
        ('[.invalid]', []),
        ('[.invalid,]', []),
        ('[age]', [37]),
        (r'[a\ge]', [37]),
        ('[age,age]', [37, 37]),
        ('[age,name.first]', [37, 'Tom']),
        ('[name.first,age,invalid,invalid.invalid,name.last]', ['Tom', 37, 'Anderson']),
        ('[[age]]', [[37]]),
        ('[[age],age]', [[37], 37]),
        ('friends.0.[age,nets.#(="ig")]', [44, 'ig']),
        ('friends.0.[age,nets.#(="ig"),invalid]', [44, 'ig']),
        ('friends.0.[age,nets.#(="ig")#]', [44, ['ig']]),
        ('friends.#.[age,first]', [[44, 'Dale'], [68, 'Roger'], [47, 'Jane']]),
        ('friends.#(age>44)#.[age,first]', [[68, 'Roger'], [47, 'Jane']]),
        ('friends.#(age>44)#.[age,invalid,invalid.invalid,first]', [[68, 'Roger'], [47, 'Jane']]),
        (r'[age,name.first,fav\.movie]', [37, 'Tom', 'Deer Hunter']),
        ('[age,name.[first,last],children.@sort]', [37, ['Tom', 'Anderson'], ['Alex', 'Jack', 'Sara']]),
        ('friends.[0.first,1.last,2.age]', ['Dale', 'Craig', 47]),
        ('[friends.[0.[nets.[0]]]]', [[[['ig']]]]),
        ('[friends.[0.[nets.[0,1]]]]', [[[['ig', 'fb']]]]),
        # Multipaths mixed
        ('[{}]', [{}]),
        ('{[]}', {'_': []}),
        ('[{},[],{}]', [{}, [], {}]),
        ('{"a":[]}', {'a': []}),
        ('[{age},{name.first}]', [{'age': 37}, {'first': 'Tom'}]),
        ('{friends.0.[age,nets.#(="ig")]}', {'_': [44, 'ig']}),
        ('{friends.0.[age,nets.#(="ig")],age}', {'_': [44, 'ig'], 'age': 37}),
        ('{friends.0.[invalid,nets.#(="ig")],age,invalid}', {'_': ['ig'], 'age': 37}),
        # Literals
        ('!true', True),
        ('!false', False),
        ('!null', None),
        ('!NaN', float('nan')),
        ('!Infinity', float('inf')),
        ('!-Infinity', float('-inf')),
        ('!"key"', 'key'),
        ('!"line \\"quotes\\""', 'line "quotes"'),
        ('!"a\tb"', 'a\tb'),
        ('!0', 0),
        ('!12', 12),
        ('!-12', -12),
        ('!12.34', 12.34),
        ('!12.34E2', 1234),
        ('!12.34E+2', 1234),
        ('!12.34e-2', 0.1234),
        ('!-12.34e-2', -0.1234),
        ('friends.#.!"value"', ['value', 'value', 'value']),
        ('friends.#.!invalid', []),
        ('friends.#|!"value"', 'value'),
        ('friends.#(age>45)#.!"value"', ['value', 'value']),
        ('name|!"value"', 'value'),
        ('!{}', {}),
        ('![]', []),
        ('!{"name":{"first":"Tom"}}.{name.first}.first', 'Tom'),
        ('{name.last,"key":!"value"}', {'last': 'Anderson', 'key': 'value'}),
        ('{name.last,"key":!{"a":"b"},"invalid"}', {'last': 'Anderson', 'key': {'a': 'b'}}),
        ('{name.last,"key":!{"c":"d"},!"valid"}', {'last': 'Anderson', 'key': {'c': 'd'}, '_': 'valid'}),
        ('[!true,!false,!null,!Infinity,!invalid,{"name":!"andy",name.last},+Infinity,!["value1","value2"]]',
         [True, False, None, float('inf'), {'name': 'andy', 'last': 'Anderson'}, ['value1', 'value2']]),
        ('[!12.34,!-12.34e-2,!true]', [12.34, -0.1234, True]),
    ))
    def test_get_ok(self, query, expected):
        """It should query the JSON object and return the expected result."""
        compare_values(self.object.get(query), expected)

    @pytest.mark.parametrize('query, error', (
        # Basic
        ('.', 'Invalid query starting with a path delimiter.'),
        ('|', 'Invalid query starting with a path delimiter.'),
        ('.name', 'Invalid query starting with a path delimiter.'),
        ('|age', 'Invalid query starting with a path delimiter.'),
        ('name..first', 'Invalid query with two consecutive path delimiters.'),
        ('name||first', 'Invalid query with two consecutive path delimiters.'),
        ('name.|first', 'Invalid query with two consecutive path delimiters.'),
        ('name|.first', 'Invalid query with two consecutive path delimiters.'),
        ('age.0', "Integer query part on unsupported object type <class 'int'>"),
        ('friends.99', 'Index `99` out of range for sequence object with 3 items in query.'),
        ('name.nonexistent', 'Mapping object does not have key `nonexistent`.'),
        ('name.1', 'Mapping object does not have key `1`.'),
        ('children.invalid', 'Invalid or unsupported query part `invalid`.'),
        ('children.', 'Delimiter at the end of the query.'),
        ('children\\', 'Escape character at the end of the query.'),
        # Wildcards
        ('x*', 'No key matching pattern with wildcard `x*`'),
        ('??????????', 'No key matching pattern with wildcard `??????????`'),
        ('children.x*', "Wildcard matching key `x*` requires a mapping object, got <class 'list'> instead."),
        ('(-?', 'No key matching pattern with wildcard `(-?`'),
        # Queries
        ('#', "Expected a sequence like object for query part # at the end of the query, got <class 'dict'>."),
        ('#.invalid', 'Invalid or unsupported query part `invalid`.'),
        ('friends.#(=="Murphy")', 'Query on mapping like objects require a key before the operator.'),
        ('friends.#(last=={1: 2})', 'Invalid value `{1: 2}` for the query key `last`'),
        ('friends.#(invalid', 'Unbalanced parentheses `(`, 1 still opened.'),
        ('#(first)', 'Queries are supported only for sequence like objects'),
        ('friends.#(invalid)', 'Query for first element does not match anything.'),
        ('friends.#(last=="invalid")', 'Query for first element does not match anything.'),
        ('friends.#(first%"D?")', 'Query for first element does not match anything.'),
        ('friends.#(last=="Murphy")invalid', 'Expected delimiter or end of query after closing parenthesis.'),
        ('children.#()', 'Empty or invalid query.'),
        ('children.#()#', 'Empty or invalid query.'),
        ('friends.#.invalid.#()', 'Empty or invalid query.'),
        ('friends.#.invalid.#()#', 'Empty or invalid query.'),
        # Dot vs Pipe
        ('friends.#(last="Murphy")#|first', 'Invalid or unsupported query'),
        # Modifiers
        ('@', 'Got empty modifier name.'),
        ('friends.@', 'Got empty modifier name.'),
        ('friends.@pretty:', 'Modifier with options separator `:` without any option.'),
        ('friends.@pretty:{invalid', 'Unable to load modifier options.'),
        ('friends.@pretty:["invalid"]', "Expected JSON object `{...}` as modifier options."),
        ('friends.@invalid', 'Unknown modifier @invalid.'),
        ('friends.@in"valid', 'Invalid modifier name @in"valid, the following characters are not allowed'),
        # JSON Lines
        ('..name', 'Invalid query starting with a path delimiter.'),
        # Multipaths
        (r'{"a\ge":age}', r'Failed to parse multipaths key "a\ge"'),
        ('{"age",age}', 'Expected colon after multipaths item with key "age".'),
        ('{]', 'Unbalanced parentheses `{`, 1 still opened.'),
        ('{', 'Unbalanced parentheses `{`, 1 still opened.'),
        ('{}@pretty', 'Expected delimiter or end of query after closing parenthesis.'),
        ('[{age}}]', 'Missing separator after multipath.'),
        ('{[age]]}', 'Missing separator after multipath.'),
        ('[{age,name.first]},age]', 'Expected delimiter or end of query after closing parenthesis.'),
        # Literals
        ('!', 'Unable to load literal JSON'),
        ('name.!', 'Unable to load literal JSON'),
        ('!invalid', 'Unable to load literal JSON'),
        (r'!in\valid', 'Unable to load literal JSON'),
        ('!0.a', 'Invalid or unsupported query part `a`.'),
        ('!0.1ea', 'Invalid or unsupported query part `ea`.'),
        ('!-12.', 'Delimiter at the end of the query.'),
        ('!-12.e', 'Invalid or unsupported query part `e`.'),
        ('name.!invalid', 'Unable to load literal JSON'),
        ('!"invalid', 'Unable to find end of literal string.'),
        ('friends.#|!invalid', 'Unable to load literal JSON'),
        ('!{true,', 'Unbalanced parentheses `{`, 1 still opened.'),
        ('![true,', 'Unbalanced parentheses `[`, 1 still opened.'),
        ('!"value".invalid', 'Invalid or unsupported query part `invalid`.'),
        ('name.!"value"', 'Unable to load literal JSON: literal afer a dot delimiter.'),
    ))
    def test_get_parser_raise(self, query, error):
        """It should raise a GJSONParseError error with the expected message."""
        with pytest.raises(gjson.GJSONParseError, match=re.escape(error)):
            self.object.get(query)

    @pytest.mark.parametrize('query, error', (
        # Basic
        ('', 'Empty query.'),
        # Modifiers
        ('children.@keys', 'The current object does not have a keys() method.'),
        ('children.@values', 'The current object does not have a values() method.'),
        ('age.@group', "Modifier @group got object of type <class 'int'> as input, expected dictionary."),
        ('children.@group', "Modifier @group got object of type <class 'list'> as input, expected dictionary."),
    ))
    def test_get_raise(self, query, error):
        """It should raise a GJSONError error with the expected message."""
        with pytest.raises(gjson.GJSONError, match=re.escape(error)):
            self.object.get(query)


class TestEscape:
    """Test gjson for all the escape sequences."""

    def setup_method(self):
        """Initialize the test instance."""
        self.escape = gjson.GJSON(INPUT_ESCAPE)

    @pytest.mark.parametrize('query, expected', (
        (r'test.\*', 'valZ'),
        (r'test.\*v', 'val0'),
        (r'test.keyv\*', 'val1'),
        (r'test.key\*v', 'val2'),
        (r'test.keyv\?', 'val3'),
        (r'test.key\?v', 'val4'),
        (r'test.keyv\.', 'val5'),
        (r'test.key\.v', 'val6'),
        (r'test.keyk\*.key\?', 'val7'),
        ('test.1key', 'val8'),
    ))
    def test_get_ok(self, query, expected):
        """It should query the escape test JSON and return the expected result."""
        assert self.escape.get(query, quiet=True) == expected


class TestBasic:
    """Test gjson for basic queries."""

    def setup_method(self):
        """Initialize the test instance."""
        self.basic = gjson.GJSON(INPUT_BASIC)

    @pytest.mark.parametrize('query, expected', (
        ('loggy.programmers.#(age=101).firstName', 1002.3),
        ('loggy.programmers.#(firstName != "Brett").firstName', 'Jason'),
        ('loggy.programmers.#(firstName % "Bre*").email', 'aaaa'),
        ('loggy.programmers.#(firstName !% "Bre*").email', 'bbbb'),
        ('loggy.programmers.#(firstName == "Brett").email', 'aaaa'),
        ('loggy.programmers.#.firstName', ['Brett', 'Jason', 'Elliotte', 1002.3]),
        ('loggy.programmers.#.asd', []),
        ('items.3.tags.#', 3),
        ('items.3.points.1.#', 2),
        ('items.#', 8),
        ('vals.#', 3),
        ('name2.here', r'B\"R'),
        ('arr.#', 6),
        ('arr.3.hello', 'world'),
        ('name.first', 'tom'),
        ('name.last', None),
        ('age', 100),
        ('happy', True),
        ('immortal', False),
        ('noop', {'what is a wren?': 'a bird'}),
        # Modifiers
        ('arr.@join', {'hello': 'world'}),
    ))
    def test_get_ok(self, query, expected):
        """It should query the basic test JSON and return the expected result."""
        assert self.basic.get(query) == expected


class TestList:
    """Test gjson queries on a list object."""

    def setup_method(self):
        """Initialize the test instance."""
        self.list = gjson.GJSON(INPUT_LIST)

    @pytest.mark.parametrize('query, expected', (
        # Dot vs Pipe
        ('#.first', ['Dale', 'Jane']),
        ('#.first.#', []),
        ('#.first|#', 2),
        ('#.0', []),
        ('#.#', []),
        # Queries
        ('#(first)#', [{'first': 'Dale'}, {'first': 'Jane'}]),
        ('#(first)', {'first': 'Dale'}),
        ('#(last)#', [{'last': 'Murphy'}]),
        ('#(last)', {'last': 'Murphy'}),
        # Modifiers
        ('@join', {'first': 'Jane', 'last': 'Murphy'}),
        # Multipaths
        ('#.{first.@reverse}', [{'@reverse': 'Dale'}, {'@reverse': 'Jane'}, {}]),
    ))
    def test_get_ok(self, query, expected):
        """It should query the list test JSON and return the expected result."""
        assert self.list.get(query, quiet=False) == expected

    @pytest.mark.parametrize('query, error', (
        # Dot vs Pipe
        ('#|first', 'Invalid or unsupported query part `first`.'),
        ('#|0', 'Integer query part after a pipe delimiter on an sequence like object.'),
        ('#|#', 'The pipe delimiter cannot immediately follow the # element.'),
    ))
    def test_get_raise(self, query, error):
        """It should raise a GJSONError error with the expected message."""
        with pytest.raises(gjson.GJSONParseError, match=re.escape(error)):
            self.list.get(query)


class TestFlattenModifier:
    """Test gjson @flatten modifier."""

    def setup_method(self):
        """Initialize the test instance."""
        self.list = gjson.GJSON(json.loads('[1, [2], [3, 4], [5, [6, 7]], [8, [9, [10, 11]]]]'))

    @pytest.mark.parametrize('query, expected', (
        ('@flatten', [1, 2, 3, 4, 5, [6, 7], 8, [9, [10, 11]]]),
        ('@flatten:{"deep":true}', [1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11]),
    ))
    def test_get(self, query, expected):
        """It should correctly flatten the given object."""
        assert self.list.get(query, quiet=True) == expected


class TestTruthiness:
    """Testing gjson with an input object with truthy/falsy objects."""

    def setup_method(self):
        """Initialize the test instance."""
        self.object = gjson.GJSON(INPUT_TRUTHINESS)

    @pytest.mark.parametrize('query, expected', (
        ('vals.#(b==~true).a', 1),
        ('vals.#(b==~true)#.a', [1, 2, 4, 6, 7, 8]),
        ('vals.#(b==~false).a', 3),
        ('vals.#(b==~false)#.a', [3, 5, 9, 10, 11]),
        ('vals.#(b==~"invalid")#', []),
    ))
    def test_get_ok(self, query, expected):
        """It should query the JSON object and return the expected result."""
        compare_values(self.object.get(query), expected)

    @pytest.mark.parametrize('query, error', (
        ('vals.#(b==~"invalid")',
         "Queries ==~ operator requires a boolean value, got <class 'str'> instead: `invalid`"),
    ))
    def test_get_raise(self, query, error):
        """It should raise a GJSONError error with the expected message."""
        with pytest.raises(gjson.GJSONError, match=re.escape(error)):
            self.object.get(query)


class TestNestedQueries:
    """Testing gjson nested queries."""

    def setup_method(self):
        """Initialize the test instance."""
        self.object = gjson.GJSON(INPUT_NESTED_QUERIES)

    @pytest.mark.parametrize('query, expected', (
        # Arrays of objects
        ('key.#(level1.#(level2.#(level3)))', INPUT_NESTED_QUERIES['key'][0]),
        ('key.#(level1.#(level2.#(level3)))#', INPUT_NESTED_QUERIES['key'][0:2]),
        ('key.#(level1.#(level2.#(level3.#(==0))))#', []),
        ('key.#(level1.#(level2.#(level3.#(=1))))', INPUT_NESTED_QUERIES['key'][0]),
        ('key.#(level1.#(level2.#(level3.#(=1)#)#)#)', INPUT_NESTED_QUERIES['key'][0]),
        ('key.#(level1.#(level2.#(level3.#(==1))))#', [INPUT_NESTED_QUERIES['key'][0]]),
        ('key.#(level1.#(level2.#(level3.#(==2))))', INPUT_NESTED_QUERIES['key'][0]),
        ('key.#(level1.#(level2.#(level3.#(=2))))#', INPUT_NESTED_QUERIES['key'][0:2]),
        # Arrays of arrays
        ('key.#(#(#(level3)))', INPUT_NESTED_QUERIES['key'][2]),
        ('key.#(#(#(level3)))#', INPUT_NESTED_QUERIES['key'][2:4]),
        ('key.#(#(#(level3.#(==0))))#', []),
        ('key.#(#(#(level3.#(==1))))', INPUT_NESTED_QUERIES['key'][2]),
        ('key.#(#(#(level3.#(==1)#)#)#)', INPUT_NESTED_QUERIES['key'][2]),
        ('key.#(#(#(level3.#(==1))))#', [INPUT_NESTED_QUERIES['key'][2]]),
        ('key.#(#(#(level3.#(==2))))', INPUT_NESTED_QUERIES['key'][2]),
        ('key.#(#(#(level3.#(==2))))#', INPUT_NESTED_QUERIES['key'][2:4]),
        ('key.#(#(#(level3.#(>=4))))', INPUT_NESTED_QUERIES['key'][3]),
        ('key.#(#(#(level3.#(>=4))))#', [INPUT_NESTED_QUERIES['key'][3]]),
        # Mixed
        ('key.#(mixed.#(#(level4)))', INPUT_NESTED_QUERIES['key'][-1]),
        ('key.#(mixed.#(#(level4)))#', [INPUT_NESTED_QUERIES['key'][-1]]),
    ))
    def test_get_ok(self, query, expected):
        """It should query the JSON object and return the expected result."""
        compare_values(self.object.get(query), expected)

    @pytest.mark.parametrize('query, error', (
        ('key.#(level1.#(level2.#(level3.#(==0))))', 'Query for first element does not match anything.'),
        ('key.#(#(#(level3.#(==0))))', 'Query for first element does not match anything.'),
    ))
    def test_get_raise(self, query, error):
        """It should raise a GJSONError error with the expected message."""
        with pytest.raises(gjson.GJSONError, match=re.escape(error)):
            self.object.get(query)


@pytest.mark.parametrize('query, expected', (
    ('0.0', 'zero'),
    ('0|0', 'zero'),
    ('#.0', ['zero']),
    ('#.1', ['one', 'one']),
    ('#.9', []),
    ('#(0="zero")#|0', {'0': 'zero', '1': 'one'}),
    ('#(0="zero")#.1', ['one']),
    ('#(0="zero")#.9', []),
    ('#(0="invalid")#.1', []),
))
def test_get_integer_mapping_keys_ok(query, expected):
    """It should return the expected result."""
    obj = gjson.GJSON([{'0': 'zero', '1': 'one'}, {'1': 'one'}])
    assert obj.get(query, quiet=True) == expected


@pytest.mark.parametrize('query, error', (
    ('0.1', 'Mapping object does not have key `1`.'),
    ('#|0', 'Integer query part after a pipe delimiter on an sequence like object.'),
    ('#|9', 'Integer query part after a pipe delimiter on an sequence like object.'),
    ('#(0="zero")#|1', 'Index `1` out of range for sequence object with 1 items in query.'),
))
def test_get_integer_mapping_keys_raise(query, error):
    """It should return the expected result."""
    with pytest.raises(gjson.GJSONError, match=re.escape(error)):
        gjson.GJSON([{'0': 'zero'}]).get(query)


@pytest.mark.parametrize('modifier', ('@valid', '@this'))
def test_get_modifier_unmodified_ok(modifier):
    """It should return the same object."""
    obj = gjson.GJSON(INPUT_OBJECT)
    assert obj.get(modifier, quiet=True) == INPUT_OBJECT


def test_get_modifier_valid_raise():
    """It should return None if the object is invalid JSON and quiet is True."""
    obj = gjson.GJSON({'invalid': {1, 2}})
    assert obj.get('@valid', quiet=True) is None


@pytest.mark.parametrize('data, expected', (
    ('[3, 1, 5, 8, 2]', [1, 2, 3, 5, 8]),
    ('{"b": 2, "d": 4, "c": 3, "a": 1}', {"a": 1, "b": 2, "c": 3, "d": 4}),
    ('"a string"', None),
))
def test_get_modifier_sort(data, expected):
    """It should return the object sorted."""
    obj = gjson.GJSON(json.loads(data))
    compare_values(obj.get('@sort', quiet=True), expected)


@pytest.mark.parametrize('data, query, expected', (
    ({'a': '{"b": 25}'}, 'a.@fromstr', {'b': 25}),
    ({'a': '{"b": 25}'}, 'a.@fromstr.b', 25),
))
def test_get_modifier_fromstr_ok(data, query, expected):
    """It should load the JSON-encoded string."""
    obj = gjson.GJSON(data)
    assert obj.get(query, quiet=True) == expected


@pytest.mark.parametrize('query, error', (
    ('a.@fromstr', 'The current @fromstr input object cannot be converted to JSON.'),
    ('b.@fromstr', "Modifier @fromstr got object of type <class 'dict'> as input, expected string or bytes."),
))
def test_get_modifier_fromstr_raise(query, error):
    """It should raise a GJSONError if the JSON-encoded string has invalid JSON."""
    obj = gjson.GJSON({'a': '{"invalid: json"', 'b': {'not': 'a string'}})
    with pytest.raises(gjson.GJSONError, match=re.escape(error)):
        obj.get(query)


def test_get_modifier_tostr_raise():
    """It should raise a GJSONError if the object cannot be JSON-encoded."""
    obj = gjson.GJSON({'a': {1, 2, 3}})  # Python sets cannot be JSON-encoded
    match = re.escape('The current object cannot be converted to a JSON-encoded string for @tostr.')
    with pytest.raises(gjson.GJSONError, match=match):
        obj.get('a.@tostr')


def test_get_modifier_group_ok():
    """It should group the dict of lists into a list of dicts."""
    obj = gjson.GJSON({
        'invalid1': 5,
        'id': ['123', '456', '789'],
        'val': [2, 1],
        'invalid2': 'invalid',
        'unit': ['ms', 's', 's', 'ms'],
    })
    assert obj.get('@group') == [
        {'id': '123', 'val': 2, 'unit': 'ms'},
        {'id': '456', 'val': 1, 'unit': 's'},
        {'id': '789', 'unit': 's'},
        {'unit': 'ms'},
    ]


def test_get_modifier_group_empty():
    """It should return an empty list if no values are lists or are empty."""
    obj = gjson.GJSON({'invalid1': 5, 'invalid2': 'invalid', 'invalid3': {'a': 5}, 'id': []})
    assert obj.get('@group') == []


def test_get_integer_index_on_mapping():
    """It should access the integer as string key correctly."""
    obj = gjson.GJSON(json.loads('{"1": 5, "11": 7}'))
    assert obj.get('1') == 5
    assert obj.get('11') == 7


def test_module_get():
    """It should return the queried object."""
    assert gjson.get({'key': 'value'}, 'key') == 'value'


def test_gjson_get_gjson():
    """It should return the queried object as a GJSON object."""
    ret = gjson.GJSON(INPUT_OBJECT).get_gjson('children')
    assert isinstance(ret, gjson.GJSON)
    assert str(ret) == '["Sara", "Alex", "Jack"]'


@pytest.mark.parametrize('data, num, expected', (
    # Valid data
    ('[1, 2, 3, 4, 5]', None, {1: 1, 2: 1, 3: 1, 4: 1, 5: 1}),
    ('[1, 2, 3, 4, 5]', 0, {}),
    ('[1, 2, 3, 4, 5]', 2, {1: 1, 2: 1}),
    ('[1, 1, 1, 1, 1]', None, {1: 5}),
    ('[1, 1, 1, 1, 1]', 1, {1: 5}),
    ('[1, 1, 1, 2, 2, 3]', None, {1: 3, 2: 2, 3: 1}),
    ('[1, 1, 1, 2, 2, 3, 3, 3, 3]', None, {3: 4, 1: 3, 2: 2}),
    ('[1, 1, 1, 2, 2, 3, 3, 3, 3]', 2, {3: 4, 1: 3}),
    # Invalid data
    ('{"key": "value"}', None, None),
    ('1', None, None),
    ('"value"', None, None),
))
def test_get_modifier_top_n(data, num, expected):
    """It should return the top N common items."""
    obj = gjson.GJSON(json.loads(data))
    if num is not None:
        compare_values(obj.get(f'@top_n:{{"n": {num}}}', quiet=True), expected)
    else:
        compare_values(obj.get('@top_n', quiet=True), expected)


@pytest.mark.parametrize('num, expected', (
    (0, {}),
    (1, {"c": 12}),
    (2, {"c": 12, "a": 8}),
    (3, {"c": 12, "a": 8, "d": 4}),
    (4, {"c": 12, "a": 8, "d": 4, "b": 3.5}),
    (None, {"c": 12, "a": 8, "d": 4, "b": 3.5}),
))
def test_get_modifier_sum_n_valid(num, expected):
    """It should group and sum and return the top N items."""
    obj = gjson.GJSON(INPUT_SUM_N)
    if num is not None:
        compare_values(obj.get(f'@sum_n:{{"group": "key", "sum": "value", "n": {num}}}', quiet=True), expected)
    else:
        compare_values(obj.get('@sum_n:{"group": "key", "sum": "value"}', quiet=True), expected)


@pytest.mark.parametrize('data', (
    '{"an": "object"}',
    '"a string"',
    '1',
))
def test_get_modifier_sum_n_invalid_data(data):
    """It should raise a GJSONError if the input is invalid."""
    obj = gjson.GJSON(json.loads(data))
    with pytest.raises(gjson.GJSONError, match="@sum_n modifier not supported for object of type"):
        obj.get('@sum_n:{"group": "key", "sum": "value"}')


@pytest.mark.parametrize('options', (
    '',
    ':{}',
    ':{"group": "invalid", "sum": "value"}',
    ':{"group": "key", "sum": "invalid"}',
    ':{"group": "key", "sum": "other"}',
    ':{"group": "other", "sum": "value"}',
))
def test_get_modifier_sum_n_invalid_options(options):
    """It should raise a GJSONError if the options are invalid."""
    obj = gjson.GJSON(INPUT_SUM_N)
    with pytest.raises(gjson.GJSONError, match="Modifier @sum_n raised an exception"):
        obj.get(f'@sum_n{options}')


class TestJSONOutput:
    """Test class for all JSON output functionalities."""

    def setup_method(self):
        """Initialize the test instance."""
        self.obj = {'key': 'value', 'hello world': '\u3053\u3093\u306b\u3061\u306f\u4e16\u754c'}
        self.query = 'key'
        self.value = '"value"'
        self.gjson = gjson.GJSON(self.obj)

    def test_module_get_as_str(self):
        """It should return the queried object as a JSON string."""
        assert gjson.get(self.obj, self.query, as_str=True) == self.value
        assert gjson.get(self.obj, '', as_str=True, quiet=True) == ''

    def test_gjson_getj(self):
        """It should return the queried object as a JSON string."""
        assert self.gjson.getj(self.query) == self.value
        assert self.gjson.getj('', quiet=True) == ''

    def test_module_get_as_str_raise(self):
        """It should raise a GJSONError with the proper message on failure."""
        with pytest.raises(gjson.GJSONError, match='Empty query.'):
            gjson.get(self.obj, '', as_str=True)

    def test_gjson_get_as_str_raise(self):
        """It should raise a GJSONError with the proper message on failure."""
        with pytest.raises(gjson.GJSONError, match='Empty query.'):
            self.gjson.getj('')

    @pytest.mark.parametrize('query, expected', (
        ('@pretty', '{\n  "key": "value",\n  "hello world": "\u3053\u3093\u306b\u3061\u306f\u4e16\u754c"\n}'),
        ('@pretty:{"indent": 4}',
         '{\n    "key": "value",\n    "hello world": "\u3053\u3093\u306b\u3061\u306f\u4e16\u754c"\n}'),
        ('@pretty:{"indent": "\t"}',
         '{\n\t"key": "value",\n\t"hello world": "\u3053\u3093\u306b\u3061\u306f\u4e16\u754c"\n}'),
        # Multipaths
        ('{key,"another":key}.@pretty', '{\n  "key": "value",\n  "another": "value"\n}'),
    ))
    def test_modifier_pretty(self, query, expected):
        """It should prettyfy the JSON string based on the parameters."""
        assert self.gjson.getj(query) == expected

    def test_modifier_pretty_sort_keys_prefix(self):
        """It should prettyfy the JSON string and sort the keys."""
        output = gjson.GJSON({'key2': 'value2', 'key1': 'value1'}).getj('@pretty:{"sortKeys": true, "prefix": "## "}')
        assert output == '## {\n##   "key1": "value1",\n##   "key2": "value2"\n## }'

    def test_modifier_ugly(self):
        """It should uglyfy the JSON string."""
        assert gjson.get(self.obj, '@ugly', as_str=True) == (
            '{"key":"value","hello world":"\u3053\u3093\u306b\u3061\u306f\u4e16\u754c"}')

    def test_output_unicode(self):
        """It should return unicode characters as-is."""
        assert gjson.get(self.obj, 'hello world', as_str=True) == '"\u3053\u3093\u306b\u3061\u306f\u4e16\u754c"'

    def test_modifier_ascii(self):
        """It should escape all non-ASCII characters."""
        assert gjson.get(self.obj, 'hello world.@ascii', as_str=True) == (
            '"\\u3053\\u3093\\u306b\\u3061\\u306f\\u4e16\\u754c"')


def custom_sum(options, obj, *, last):
    """Custom modifier function."""
    assert last is True
    assert options == {}
    if not isinstance(obj, list):
        raise RuntimeError('@sum can be used only on lists')

    return sum(obj)


class TestCustomModifiers:
    """Test class for custom modifiers."""

    def setup_method(self):
        """Initialize the test instance."""
        self.valid_obj = [1, 2, 3, 4, 5]
        self.invalid_obj = 'invalid'
        self.query = '@sum'

    def test_gjson_register_modifier_ok(self):
        """It should register a valid modifier."""
        obj = gjson.GJSON(self.valid_obj)
        obj.register_modifier('sum', custom_sum)
        assert obj.get(self.query) == 15

    def test_gjson_register_modifier_with_escape_ok(self):
        """It should register a valid modifier with escaped characters in the name."""
        obj = gjson.GJSON(self.valid_obj)
        obj.register_modifier('sum\\=', custom_sum)
        assert obj.get('@sum\\=') == 15

    @pytest.mark.parametrize('char', MODIFIER_NAME_RESERVED_CHARS)
    def test_gjson_register_modifier_invalid_name(self, char):
        """It should raise a GJSONError if trying to register a modifier with a name with not allowed characters."""
        obj = gjson.GJSON(self.valid_obj)
        name = fr'a{char}b'
        with pytest.raises(
                gjson.GJSONError,
                match=fr'Unable to register modifier `{re.escape(name)}`, contains at least one not allowed'):
            obj.register_modifier(name, custom_sum)

    def test_gjson_register_modifier_override_builtin(self):
        """It should raise a GJSONError if trying to register a modifier with the same name of a built-in one."""
        obj = gjson.GJSON(self.valid_obj)
        with pytest.raises(gjson.GJSONError,
                           match=r'Unable to register a modifier with the same name of the built-in modifier: @valid'):
            obj.register_modifier('valid', custom_sum)

    def test_gjson_register_modifier_not_callable(self):
        """It should raise a GJSONError if trying to register a modifier that is not callable."""
        obj = gjson.GJSON(self.valid_obj)
        with pytest.raises(gjson.GJSONError, match=r'The given func "not-callable" for the custom modifier @sum does'):
            obj.register_modifier('sum', 'not-callable')

    def test_gjsonobj_custom_modifiers_ok(self):
        """It should register a valid modifier."""
        obj = gjson.GJSONObj(self.valid_obj, self.query, custom_modifiers={'sum': custom_sum})
        assert obj.get() == 15

    def test_gjsonobj_custom_modifiers_raise(self):
        """It should encapsulate the modifier raised exception in a GJSONError."""
        with pytest.raises(gjson.GJSONError,
                           match=r'Modifier @sum raised an exception'):
            gjson.GJSONObj(self.invalid_obj, self.query, custom_modifiers={'sum': custom_sum}).get()

    def test_gjsonobj_custom_modifiers_override_builtin(self):
        """It should raise a GJSONError if passing custom modifiers that have the same name of a built-in one."""
        with pytest.raises(gjson.GJSONError,
                           match=r"Some provided custom_modifiers have the same name of built-in ones: {'valid'}"):
            gjson.GJSONObj(self.valid_obj, self.query, custom_modifiers={'valid': custom_sum})

    def test_gjsoniobj_custom_modifiers_not_callable(self):
        """It should raise a GJSONError if passing custom modifiers that are not callable."""
        with pytest.raises(gjson.GJSONError, match=r'The given func "not-callable" for the custom modifier @sum does'):
            gjson.GJSONObj(self.valid_obj, self.query, custom_modifiers={'sum': 'not-callable'})

    def test_gjsonobj_builtin_modifiers(self):
        """It should return a set with the names of the built-in modifiers."""
        expected = {'ascii', 'flatten', 'fromstr', 'group', 'join', 'keys', 'pretty', 'reverse', 'sort', 'sum_n',
                    'this', 'top_n', 'tostr', 'valid', 'values', 'ugly'}
        assert gjson.GJSONObj.builtin_modifiers() == expected



"""GJSON module."""
import json
import os
import re
from importlib.metadata import PackageNotFoundError, version
from typing import Any

from gjson._gjson import GJSONObj, MODIFIER_NAME_RESERVED_CHARS
from gjson._protocols import ModifierProtocol
from gjson.exceptions import GJSONError, GJSONParseError


# Explicit export of modules for the import * syntax, custom order to force the documentation order
__all__ = ['get', 'GJSON', 'GJSONError', 'GJSONParseError', 'ModifierProtocol', 'GJSONObj', '__version__']


# TODO: use a proper type hint for obj once https://github.com/python/typing/issues/182 will be fixed
def get(obj: Any, query: str, *, as_str: bool = False, quiet: bool = False) -> Any:
    """Quick accessor to GJSON functionalities exposed for simplicity of use.

    Examples:
        Import and directly use this quick helper for the simpler usage::

            >>> import gjson
            >>> data = {'items': [{'name': 'a', 'size': 1}, {'name': 'b', 'size': 2}]}
            >>> gjson.get(data, 'items.#.size')
            [1, 2]

    Arguments:
        obj: the object to query. It must be accessible in JSON-like fashion so it must be an object that can be
            converted to JSON.
        query: the query string to evaluate to extract the data from the object.
        as_str: if set to :py:data:`True` returns a JSON-encoded string, a Python object otherwise.
        quiet: on error, if set to :py:data:`True`, will raises an GJSONError exception. Otherwise returns
            :py:data:`None` on error.

    Return:
        the resulting object.

    """
    gjson_obj = GJSON(obj)
    if as_str:
        return gjson_obj.getj(query, quiet=quiet)

    return gjson_obj.get(query, quiet=quiet)


class GJSON:
    """The GJSON class to operate on JSON-like objects."""

    def __init__(self, obj: Any):
        """Initialize the instance with the given object.

        Examples:
            Use the :py:class:`gjson.GJSON` class for more complex usage or to perform multiple queries on the same
            object::

                >>> import gjson
                >>> data = {'items': [{'name': 'a', 'size': 1}, {'name': 'b', 'size': 2}]}
                >>> gjson_obj = gjson.GJSON(data)

        Arguments:
            obj: the object to query.

        """
        self._obj = obj
        self._custom_modifiers: dict[str, ModifierProtocol] = {}

    def __str__(self) -> str:
        """Return the current object as a JSON-encoded string.

        Examples:
            Converting to string a :py:class:`gjson.GJSON` object returns it as a JSON-encoded string::

                >>> str(gjson_obj)
                '{"items": [{"name": "a", "size": 1}, {"name": "b", "size": 2}]}'

        Returns:
            the JSON-encoded string representing the instantiated object.

        """
        return json.dumps(self._obj, ensure_ascii=False)

    def get(self, query: str, *, quiet: bool = False) -> Any:
        """Perform a query on the instantiated object and return the resulting object.

        Examples:
            Perform a query and get the resulting object::

                >>> gjson_obj.get('items.#.size')
                [1, 2]

        Arguments:
            query: the GJSON query to apply to the object.
            quiet: wheter to raise a :py:class:`gjson.GJSONError` exception on error or just return :py:data:`None` in
                case of error.

        Raises:
            gjson.GJSONError: on error if the quiet parameter is not :py:data:`True`.

        Returns:
            the resulting object or :py:data:`None` if the ``quiet`` parameter is :py:data:`True` and there was an
            error.

        """
        try:
            return GJSONObj(self._obj, query, custom_modifiers=self._custom_modifiers).get()
        except GJSONError:
            if quiet:
                return None
            raise

    def getj(self, query: str, *, quiet: bool = False) -> str:
        """Perform a query on the instantiated object and return the resulting object as JSON-encoded string.

        Examples:
            Perform a query and get the resulting object as a JSON-encoded string::

                >>> gjson_obj.getj('items.#.size')
                '[1, 2]'

        Arguments:
            query: the GJSON query to apply to the object.
            quiet: wheter to raise a :py:class:`gjson.GJSONError` exception on error or just return :py:data:`None` in
                case of error.

        Raises:
            gjson.GJSONError: on error if the quiet parameter is not :py:data:`True`.

        Returns:
            the JSON-encoded string representing the resulting object or :py:data:`None` if the ``quiet`` parameter is
            :py:data:`True` and there was an error.

        """
        try:
            return str(GJSONObj(self._obj, query, custom_modifiers=self._custom_modifiers))
        except GJSONError:
            if quiet:
                return ''
            raise

    def get_gjson(self, query: str, *, quiet: bool = False) -> 'GJSON':
        """Perform a query on the instantiated object and return the resulting object as a GJSON instance.

        Examples:
            Perform a query and get the resulting object already encapsulated into a :py:class:`gjson.GJSON` object::

                >>> sizes = gjson_obj.get_gjson('items.#.size')
                >>> str(sizes)
                '[1, 2]'
                >>> sizes.get('0')
                1

        Arguments:
            query: the GJSON query to apply to the object.
            quiet: wheter to raise a :py:class:`gjson.GJSONError` exception on error or just return :py:data:`None` in
                case of error.

        Raises:
            gjson.GJSONError: on error if the quiet parameter is not :py:data:`True`.

        Returns:
            the resulting object encapsulated as a :py:class:`gjson.GJSON` object or :py:data:`None` if the ``quiet``
            parameter is :py:data:`True` and there was an error.

        """
        return GJSON(self.get(query, quiet=quiet))

    def register_modifier(self, name: str, func: ModifierProtocol) -> None:
        """Register a custom modifier.

        Examples:
            Register a custom modifier that sums all the numbers in a list:

                >>> def custom_sum(options, obj, *, last):
                ...     # insert sanity checks code here
                ...     return sum(obj)
                ...
                >>> gjson_obj.register_modifier('sum', custom_sum)
                >>> gjson_obj.get('items.#.size.@sum')
                3

        Arguments:
            name: the modifier name. It will be called where ``@name`` is used in the query. If two custom modifiers
                are registered with the same name the last one will be used.
            func: the modifier code in the form of a callable object that adhere to the
                :py:class:`gjson.ModifierProtocol`.

        Raises:
            gjson.GJSONError: if the provided callable doesn't adhere to the :py:class:`gjson.ModifierProtocol`.

        """
        # Escape the ] as they are inside a [...] block
        not_allowed_regex = ''.join(MODIFIER_NAME_RESERVED_CHARS).replace(']', r'\]')
        if re.search(fr'[{not_allowed_regex}]', name):
            not_allowed_string = ', '.join(f'`{i}`' for i in MODIFIER_NAME_RESERVED_CHARS)
            raise GJSONError(f'Unable to register modifier `{name}`, contains at least one not allowed character: '
                             f'{not_allowed_string}')

        if name in GJSONObj.builtin_modifiers():
            raise GJSONError(f'Unable to register a modifier with the same name of the built-in modifier: @{name}.')

        if not isinstance(func, ModifierProtocol):
            raise GJSONError(f'The given func "{func}" for the custom modifier @{name} does not adhere '
                             'to the gjson.ModifierProtocol.')

        self._custom_modifiers[name] = func


try:
    __version__: str = version('gjson')
    """str: the version of the current gjson module."""
except PackageNotFoundError:  # pragma: no cover - this should never happen during tests
    # Read the override from the environment, if present (like inside Debian build system)
    if 'SETUPTOOLS_SCM_PRETEND_VERSION' in os.environ:
        __version__ = os.environ['SETUPTOOLS_SCM_PRETEND_VERSION']



#!/usr/bin/env python
"""Package configuration."""
from setuptools import find_packages, setup


with open('README.rst', 'r') as readme:
    long_description = readme.read()

# Extra dependencies
extras_require = {
    # Test dependencies
    'tests': [
        'bandit',
        'flake8',
        'flake8-import-order',
        'mypy',
        'pytest-cov',
        'pytest-xdist',
        'pytest',
        'sphinx_rtd_theme',
        'sphinx-argparse',
        'sphinx-autodoc-typehints',
        'Sphinx',
        'types-pkg_resources',
    ],
    'prospector': [
        'prospector[with_everything]',
        'pytest',
    ],
}

setup_requires = [
    'pytest-runner',
    'setuptools_scm',
]

setup(
    author='Riccardo Coccioli',
    author_email='volans-@users.noreply.github.com',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'Intended Audience :: End Users/Desktop',
        'Intended Audience :: Information Technology',
        'Intended Audience :: Science/Research',
        'Intended Audience :: System Administrators',
        'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Programming Language :: Python :: 3 :: Only',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: Text Processing',
        'Topic :: Utilities',
        'Typing :: Typed',
    ],
    description=('gjson-py is a Python package that provides a simple way to filter and extract data from JSON-like '
                 'objects or JSON files, using the GJSON syntax.'),
    entry_points={
        'console_scripts': [
            'gjson = gjson._cli:cli',
        ],
    },
    extras_require=extras_require,
    install_requires=[],
    keywords=['gjson', 'json'],
    license='GPLv3+',
    long_description=long_description,
    long_description_content_type='text/x-rst',
    name='gjson',
    package_data={'gjson': ['py.typed']},
    packages=find_packages(),
    platforms=['GNU/Linux', 'BSD', 'MacOSX'],
    python_requires='>=3.9',
    setup_requires=setup_requires,
    url='https://github.com/volans-/gjson-py',
    use_scm_version=True,
    zip_safe=False,
)



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

import sys

from pathlib import Path

import sphinx_rtd_theme
from pkg_resources import get_distribution

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
# import os
# import sys
# sys.path.insert(0, os.path.abspath('.'))

sys.path.insert(0, Path(__file__).resolve().parent.parent)

# -- Project information -----------------------------------------------------

project = 'gjson'
title = f'{project} Documentation'
copyright = '2022, Riccardo Coccioli'
author = 'Riccardo Coccioli'

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.
#
# The full version, including alpha/beta/rc tags.
release = get_distribution('gjson').version
# The short X.Y version.
version = release

# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.napoleon',
    'sphinx.ext.intersphinx',
    'sphinx.ext.todo',
    'sphinx.ext.coverage',
    'sphinx.ext.viewcode',
    'sphinxarg.ext',
    'sphinx_autodoc_typehints',
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = []

# -- Options for manual page output ---------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [
    ('cli', 'gjson', 'filter and extract data from JSON-like files', [author], 1),
]

# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = 'sphinx_rtd_theme'
html_theme_path = [sphinx_rtd_theme.get_html_theme_path()]

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = []

intersphinx_mapping = {
    'python': ('https://docs.python.org/3/', None),
}

# Napoleon settings
napoleon_google_docstring = True
napoleon_numpy_docstring = False
napoleon_include_init_with_doc = False
napoleon_include_private_with_doc = False
napoleon_include_special_with_doc = False
napoleon_use_admonition_for_examples = False
napoleon_use_admonition_for_notes = False
napoleon_use_admonition_for_references = False
napoleon_use_ivar = False
napoleon_use_param = True
napoleon_use_keyword = True
napoleon_use_rtype = True
napoleon_type_aliases = None
napoleon_attr_annotations = True


# Autodoc settings
autodoc_default_options = {
    # Using None as value instead of True to support the version of Sphinx used in Buster
    'members': None,
    'member-order': 'bysource',
    'show-inheritance': None,
    'special-members': '__str__,__version__,__call__',
}
autoclass_content = 'both'




"""GJSON module."""
import json
import operator
import re
from collections import Counter
from collections.abc import Callable, Mapping, Sequence
from dataclasses import dataclass
from itertools import zip_longest
from typing import Any, Optional, Union

from gjson._protocols import ModifierProtocol
from gjson.exceptions import GJSONError, GJSONInvalidSyntaxError, GJSONParseError

ESCAPE_CHARACTER = '\\'
"""str: The grammar escape character."""
DOT_DELIMITER = '.'
"""str: One of the available delimiters in the query grammar."""
PIPE_DELIMITER = '|'
"""str: One of the available delimiters in the query grammar."""
DELIMITERS = (DOT_DELIMITER, PIPE_DELIMITER)
"""tuple: All the available delimiters in the query grammar."""
MULTIPATHS_DELIMITERS = DELIMITERS + (']', '}', ',')
"""tuple: All the available delimiters in the query grammar."""
# Single character operators goes last to avoid mis-detection.
QUERIES_OPERATORS = ('==~', '==', '!=', '<=', '>=', '!%', '=', '<', '>', '%')
"""tuple: The list of supported operators inside queries."""
MODIFIER_NAME_RESERVED_CHARS = ('"', ',', '.', '|', ':', '@', '{', '}', '[', ']', '(', ')')
"""tuple: The list of reserver characters not usable in a modifier's name."""
PARENTHESES_PAIRS = {'(': ')', ')': '(', '[': ']', ']': '[', '{': '}', '}': '{'}


class NoResult:
    """A no result type to be passed around and be checked."""


@dataclass
class BaseQueryPart:
    """Base dataclass class to represent a query part."""

    start: int
    end: int
    part: str
    delimiter: str
    previous: Optional['BaseQueryPart']
    is_last: bool

    def __str__(self) -> str:
        """String representation of the part.

        Returns:
            The part property of the instance.

        """
        return self.part


class FieldQueryPart(BaseQueryPart):
    """Basic field path query part."""


class ArrayLenghtQueryPart(BaseQueryPart):
    """Hash query part, to get the size of an array."""


class ArrayIndexQueryPart(BaseQueryPart):
    """Integer query part to get an array index."""

    @property
    def index(self) -> int:
        """Return the integer representation of the query part.

        Returns:
            the index as integer.

        """
        return int(self.part)


@dataclass
class ArrayQueryQueryPart(BaseQueryPart):
    """Query part for array queries, with additional fields."""

    field: str
    operator: str
    value: Union[str, 'ArrayQueryQueryPart']
    first_only: bool


@dataclass
class ModifierQueryPart(BaseQueryPart):
    """Modifier query part."""

    name: str
    options: dict[Any, Any]


@dataclass
class MultipathsItem:
    """Single multipaths query item."""

    key: str
    values: list[BaseQueryPart]


@dataclass
class MultipathsObjectQueryPart(BaseQueryPart):
    """JSON object multipaths query part."""

    parts: list[MultipathsItem]


@dataclass
class MultipathsArrayQueryPart(BaseQueryPart):
    """JSON object multipaths query part."""

    parts: list[list[BaseQueryPart]]


class LiteralQueryPart(BaseQueryPart):
    """Literal query part."""


class GJSONObj:
    """A low-level class to perform the GJSON query on a JSON-like object."""

    # pylint: disable=too-many-branches,too-many-statements,too-many-locals

    def __init__(self, obj: Any, query: str, *, custom_modifiers: Optional[dict[str, ModifierProtocol]] = None):
        """Initialize the instance with the starting object and query.

        Examples:
            Client code should not need to instantiate this low-level class in normal circumastances::

                >>> import gjson
                >>> data = {'items': [{'name': 'a', 'size': 1}, {'name': 'b', 'size': 2}]}
                >>> gjson_obj = gjson.GJSONObj(data, 'items.#.size')

        Arguments:
            obj: the JSON-like object to query.
            query: the GJSON query to apply to the object.
            custom_modifiers: an optional dictionary with the custom modifiers to load. The dictionary keys are the
                names of the modifiers and the values are the callables with the modifier code that adhere to the
                :py:class:`gjson.ModifierProtocol` protocol.

        Raises:
            gjson.GJSONError: if any provided custom modifier overrides a built-in one or is not callable.

        """
        self._obj = obj
        self._query = query
        if custom_modifiers is not None:
            if (intersection := self.builtin_modifiers().intersection(set(custom_modifiers.keys()))):
                raise GJSONError(f'Some provided custom_modifiers have the same name of built-in ones: {intersection}.')

            for name, modifier in custom_modifiers.items():
                if not isinstance(modifier, ModifierProtocol):
                    raise GJSONError(f'The given func "{modifier}" for the custom modifier @{name} does not adhere '
                                     'to the gjson.ModifierProtocol.')

        self._custom_modifiers = custom_modifiers if custom_modifiers else {}
        self._dump_params: dict[str, Any] = {'ensure_ascii': False}
        self._after_hash = False
        self._after_query_all = False

    @classmethod
    def builtin_modifiers(cls) -> set[str]:
        """Return the names of the built-in modifiers.

        Returns:
            the names of the built-in modifiers.

        """
        prefix = '_apply_modifier_'
        return {modifier[len(prefix):] for modifier in dir(cls) if modifier.startswith(prefix)}

    def get(self) -> Any:
        """Perform the query and return the resulting object.

        Examples:
            Returns the resulting object::

                >>> gjson_obj.get()
                [1, 2]

        Raises:
            gjson.GJSONError: on error.

        Returns:
            the resulting object.

        """
        # Reset internal parameters
        self._dump_params = {'ensure_ascii': False}
        self._after_hash = False
        self._after_query_all = False

        if not self._query:
            raise GJSONError('Empty query.')

        obj = self._obj
        for part in self._parse(start=0, end=len(self._query) - 1):
            obj = self._parse_part(part, obj)

        return obj

    def __str__(self) -> str:
        """Return the JSON string representation of the object, based on the query parameters.

        Examples:
            Returns the resulting object as a JSON-encoded string::

                >>> str(gjson_obj)
                '[1, 2]'

        Raises:
            gjson.GJSONError: on error.

        Returns:
            the JSON encoded string.

        """
        obj = self.get()
        prefix = self._dump_params.pop('prefix', '')
        json_string = json.dumps(obj, **self._dump_params)

        if prefix:
            json_string = '\n'.join(f'{prefix}{line}' for line in json_string.splitlines())

        return json_string

    def _parse(self, *, start: int, end: int, max_end: int = 0, delimiter: str = '',
               in_multipaths: bool = False) -> list[BaseQueryPart]:
        """Main parser of the query that will delegate to more specific parsers for each different feature.

        Arguments:
            start: the start position in the query.
            end: the end position in the query.
            max_end: an optional last position up to where a closing parentheses can be searched.
            delimiter: the optional delimiter before the query, if this is called on a multipaths.
            in_multipaths: whether the part to be parsed is inside a multipaths.

        Raises:
            gjson.GJSONParseError: on error.

        Returns:
            the resulting object.

        """
        current: list[str] = []
        current_start = -1
        parts: list[BaseQueryPart] = []
        previous: Optional[BaseQueryPart] = None
        require_delimiter = False

        i = start
        while True:
            part: Optional[BaseQueryPart] = None
            # Get current and next character in the query
            if i == end:
                next_char = None
            elif i >= end:
                if parts and not current:
                    parts[-1].is_last = True
                break
            else:
                next_char = self._query[i + 1]

            char = self._query[i]

            if char in DELIMITERS:
                if i == start:
                    raise GJSONParseError('Invalid query starting with a path delimiter.',
                                          query=self._query, position=i)

                if next_char in DELIMITERS:
                    raise GJSONParseError('Invalid query with two consecutive path delimiters.',
                                          query=self._query, position=i)
                if current:
                    part = FieldQueryPart(start=current_start, end=i - 1, part=''.join(current),
                                          delimiter=delimiter, previous=previous, is_last=False)
                    parts.append(part)
                    previous = part
                    current = []
                    current_start = -1

                delimiter = char
                require_delimiter = False
                if next_char is None:
                    raise GJSONParseError('Delimiter at the end of the query.', query=self._query, position=i)

                i += 1
                continue

            if char == '@':
                part = self._parse_modifier_query_part(i, delimiter, max_end=max_end, in_multipaths=in_multipaths)
            elif char == '#' and (next_char in DELIMITERS or next_char is None):
                part = ArrayLenghtQueryPart(start=i, end=i, part=char, delimiter=delimiter, is_last=next_char is None,
                                            previous=previous)
            elif char == '#' and next_char == '(':
                part = self._parse_array_query_query_part(i, delimiter, max_end=max_end)
            elif re.match(r'[0-9]', char) and not current:
                part = self._parse_array_index_query_part(i, delimiter, in_multipaths=in_multipaths)
            elif char == '{':
                part = self._parse_object_multipaths_query_part(i, delimiter, max_end=max_end)
                require_delimiter = True
            elif char == '[':
                part = self._parse_array_multipaths_query_part(i, delimiter, max_end=max_end)
                require_delimiter = True
            elif char == '!':
                part = self._parse_literal_query_part(i, delimiter, max_end=max_end, in_multipaths=in_multipaths)
            elif in_multipaths and char == ',':
                i -= 1
                break
            elif in_multipaths and require_delimiter:
                raise GJSONInvalidSyntaxError('Missing separator after multipath.', query=self._query, position=i)

            if part:
                part.previous = previous
                parts.append(part)
                previous = part
            else:  # Normal path, no special grammar
                if not current:
                    current_start = i

                current.append(char)
                if char == ESCAPE_CHARACTER:
                    i += 1  # Skip the escaped character
                    if next_char is None:
                        raise GJSONParseError('Escape character at the end of the query.',
                                              query=self._query, position=i)

                    current.append(next_char)

            if part:
                i = part.end + 1
            else:
                i += 1

        if part is None and current:
            part = FieldQueryPart(start=current_start, end=i, part=''.join(current),
                                  delimiter=delimiter, previous=previous, is_last=True)
            parts.append(part)

        return parts

    @staticmethod
    def _is_sequence(obj: Any) -> bool:
        """Check if an object is a sequence but not a string or bytes object.

        Arguments:
            obj: the object to test.

        Returns:
            :py:data:`True` if the object is a sequence but not a string or bytes, :py:data:`False` otherwise.

        """
        return isinstance(obj, Sequence) and not isinstance(obj, (str, bytes))

    def _find_closing_parentheses(self, *, start: int, opening: str, suffix: str = '', max_end: int = 0) -> int:
        """Find the matching parentheses that closes the opening one looking for unbalance of the given character.

        Arguments:
            start: the index of the opening parentheses in the query.
            opening: the opening parentheses to look for imbalances.
            suffix: an optional suffix that can be present after the closing parentheses before reaching a delimiter or
                the end of the query.
            max_end: an optional last position up to where the parentheses can be found.

        Raises:
            gjson.GJSONParseError: if unable to find the closing parentheses or the parentheses are not balanced.

        Returns:
            the position of the closing parentheses if there is no suffix or the one of the last character of the
            suffix if present.

        """
        closing = PARENTHESES_PAIRS[opening]
        opened = 0
        end = -1
        escaped = False
        in_string = False
        query = self._query[start:max_end + 1] if max_end else self._query[start:]

        for i, char in enumerate(query):
            if char == ESCAPE_CHARACTER:
                escaped = True
                continue

            if escaped:
                escaped = False
                continue

            if char == '"':
                if in_string:
                    in_string = False
                else:
                    in_string = True

                continue

            if in_string:
                continue

            if char == opening:
                opened += 1
            elif char == closing:
                if opened:
                    opened -= 1
                    if not opened:
                        end = i
                        break

        if opened or end < 0:
            raise GJSONParseError(f'Unbalanced parentheses `{opening}`, {opened} still opened.',
                                  query=self._query, position=start)

        if suffix and end + len(suffix) < len(query) and query[end + 1:end + len(suffix) + 1] == suffix:
            end += len(suffix)

        if end + 1 < len(query):
            delimiters = list(MULTIPATHS_DELIMITERS) if max_end else list(DELIMITERS)
            if opening == '(' and suffix == '#':  # Nested queries
                delimiters.append(')')

            if (max_end and query[end + 1] not in delimiters) or (not max_end and query[end + 1] not in DELIMITERS):
                raise GJSONParseError('Expected delimiter or end of query after closing parenthesis.',
                                      query=self._query, position=start + end)

        return start + end

    def _parse_modifier_query_part(self, start: int, delimiter: str, max_end: int = 0,
                                   in_multipaths: bool = False) -> ModifierQueryPart:
        """Find the modifier end position in the query starting from a given point.

        Arguments:
            start: the index of the ``@`` symbol that starts a modifier in the query.
            delimiter: the delimiter before the modifier.
            max_end: an optional last position up to where the last character can be found.
            in_multipaths: whether the part to be parsed is inside a multipaths.

        Raises:
            gjson.GJSONParseError: on invalid modifier.

        Returns:
            the modifier query part object.

        """
        end = start
        escaped = False
        delimiters = MULTIPATHS_DELIMITERS if in_multipaths else DELIMITERS
        options: dict[Any, Any] = {}
        query = self._query[start:max_end + 1] if max_end else self._query[start:]
        for i, char in enumerate(query):
            if char == ESCAPE_CHARACTER and not escaped:
                escaped = True
                continue

            if escaped:
                escaped = False
                continue

            if char == ':':
                name = self._query[start + 1:start + i]
                options_len, options = self._parse_modifier_options(start + i + 1)
                end = start + i + options_len
                break

            if char in delimiters:
                end = start + i - 1
                name = self._query[start + 1:start + i]
                break

        else:  # End of query
            end = start + i
            name = self._query[start + 1:start + i + 1]

        name.replace(ESCAPE_CHARACTER, '')
        if not name:
            raise GJSONParseError('Got empty modifier name.', query=self._query, position=start)

        for char in MODIFIER_NAME_RESERVED_CHARS:
            if char in name:
                raise GJSONParseError(f'Invalid modifier name @{name}, the following characters are not allowed: '
                                      f'{MODIFIER_NAME_RESERVED_CHARS}', query=self._query, position=start)

        return ModifierQueryPart(start=start, end=end, part=self._query[start:end + 1], delimiter=delimiter,
                                 name=name, options=options, is_last=False, previous=None)

    def _parse_modifier_options(self, start: int) -> tuple[int, dict[Any, Any]]:
        """Find the modifier options end position in the query starting from a given point.

        Arguments:
            start: the index of the ``:`` symbol that starts a modifier options.

        Raises:
            gjson.GJSONParseError: on invalid modifier options.

        Returns:
            the modifier options last character index in the query and the parsed options.

        """
        if start >= len(self._query):
            raise GJSONParseError('Modifier with options separator `:` without any option.',
                                  query=self._query, position=start)

        if self._query[start] != '{':
            raise GJSONParseError('Expected JSON object `{...}` as modifier options.',
                                  query=self._query, position=start)

        query_parts = re.split(r'(?<!\\)(\.|\|)', self._query[start:])
        options = None
        options_string = ''
        for i in range(0, len(query_parts), 2):
            options_string = ''.join(query_parts[:i + 1])
            try:
                options = json.loads(options_string, strict=False)
                break
            except json.JSONDecodeError:
                pass
        else:
            raise GJSONParseError('Unable to load modifier options.', query=self._query, position=start)

        return len(options_string), options

    def _parse_array_query_query_part(self, start: int, delimiter: str, max_end: int = 0) -> ArrayQueryQueryPart:
        """Parse an array query part starting from the given point.

        Arguments:
            start: the index of the ``#`` symbol that starts a ``#(...)`` or ``#(...)#`` query.
            delimiter: the delimiter before the modifier.
            max_end: an optional last position up to where the closing parentheses can be found.

        Raises:
            gjson.GJSONParseError: on invalid query.

        Returns:
            the array query part object.

        """
        end = self._find_closing_parentheses(start=start, opening='(', suffix='#', max_end=max_end)
        part = self._query[start:end + 1]
        if part[-1] == '#':
            content_end = -2
            first_only = False
        else:
            content_end = -1
            first_only = True

        content = part[2: content_end]
        query_operator = ''
        key = ''
        value: Union[str, ArrayQueryQueryPart] = ''

        pattern = '|'.join(re.escape(op) for op in QUERIES_OPERATORS)
        match = re.search(fr'(?<!\\)({pattern}|\.?#\()', content)  # Negative lookbehind assertion
        if match:
            query_operator = match.group()
            key = content[:match.start()]
            value = content[match.end():].strip()
        else:  # No operator, assume existence match of key
            key = content

        key = key.strip()
        if match and '#(' in query_operator:  # Nested queries
            offset = 1 if query_operator[0] == '.' else 0
            query_operator = ''
            nested_start = start + 2 + match.start() + offset
            value = self._parse_array_query_query_part(nested_start, delimiter, max_end=end + content_end)
            value.first_only = False  # Nested queries first_only is controlled by the most external query

        if not key and not (query_operator and value) and not isinstance(value, ArrayQueryQueryPart):
            raise GJSONParseError('Empty or invalid query.', query=self._query, position=start)

        return ArrayQueryQueryPart(start=start, end=end, part=part, delimiter=delimiter, is_last=False, previous=None,
                                   field=key, operator=query_operator, value=value, first_only=first_only)

    def _parse_array_index_query_part(self, start: int, delimiter: str,
                                      in_multipaths: bool = False) -> Optional[ArrayIndexQueryPart]:
        """Parse an array index query part.

        Arguments:
            start: the index of the start of the path in the query.
            delimiter: the delimiter before the query part.
            in_multipaths: whether the part to be parsed is inside a multipaths.

        Returns:
            the array index query object if the integer path is found, :py:const:`None` otherwise.

        """
        subquery = self._query[start:]
        delimiters = MULTIPATHS_DELIMITERS if in_multipaths else DELIMITERS
        delimiters_match = '|'.join([re.escape(i) for i in delimiters])
        match = re.search(fr'^([1-9][0-9]*|0)({delimiters_match}|$)', subquery)
        if not match:
            return None

        end = start + len(match.groups()[0]) - 1
        part = self._query[start:end + 1]

        return ArrayIndexQueryPart(start=start, end=end, part=part, delimiter=delimiter, is_last=False, previous=None)

    def _parse_object_multipaths_query_part(
            self, start: int, delimiter: str, max_end: int = 0) -> MultipathsObjectQueryPart:
        """Parse a multipaths object query part.

        Arguments:
            start: the index of the start of the path in the query.
            delimiter: the delimiter before the query part.
            max_end: an optional last position up to where the multipaths can extend.

        Returns:
            the multipaths object query part.

        Raises:
            gjson.GJSONParseError: on invalid query.

        """
        end = self._find_closing_parentheses(start=start, opening='{', max_end=max_end)
        part = self._query[start:end + 1]
        parts = []

        def _get_key(current_key: Optional[str], value: Optional[BaseQueryPart]) -> str:
            """Return the current key or the default value if not set. Allow for empty key as valid key.

            Arguments:
                current_key: the current key to evaluate.
                value: the current value from where to extract a key name if missing.

            """
            if current_key is not None:
                return current_key

            if value and isinstance(value, (FieldQueryPart, ArrayIndexQueryPart, ModifierQueryPart)):
                return value.part

            return '_'

        new_item = True
        escaped = False
        key: Optional[str] = None
        key_start = 0
        value_start = 0
        skip_until = 0

        for i, char in enumerate(part[1:-1], start=1):
            if skip_until and i <= skip_until:
                if i == skip_until:
                    skip_until = 0
                    new_item = True

                continue

            if new_item:
                value_start = 0
                if char == '"':
                    key_start = i
                    new_item = False
                    continue

                if char != ',':
                    value_start = i
                    new_item = False

            if key_start:
                if char == ESCAPE_CHARACTER and not escaped:
                    escaped = True
                    continue

                if escaped:
                    escaped = False
                    continue

                if char == '"':
                    try:
                        key = json.loads(part[key_start:i + 1], strict=False)
                    except json.JSONDecodeError as ex:
                        raise GJSONParseError(f'Failed to parse multipaths key {part[key_start:i + 1]}.',
                                              query=self._query, position=key_start) from ex
                    key_start = 0
                    continue

            if key is not None and not key_start and not value_start:
                if char == ':':
                    value_start = i + 1
                    continue

                raise GJSONParseError(f'Expected colon after multipaths item with key "{key}".',
                                      query=self._query, position=i)

            if value_start:
                try:
                    values = self._parse(
                        start=start + value_start,
                        end=end - 1,
                        max_end=max_end - 1 if max_end else end - 1,
                        delimiter=delimiter,
                        in_multipaths=True)
                except GJSONInvalidSyntaxError:
                    raise
                except GJSONParseError:  # In multipaths, paths that fails are silently suppressed
                    values = []

                if values:
                    parts.append(MultipathsItem(key=_get_key(key, values[-1]), values=values))
                    skip_until = values[-1].end - start + 1
                else:
                    skip_until = end - start

                new_item = True
                key = None
                key_start = 0
                value_start = 0
                continue

        return MultipathsObjectQueryPart(start=start, end=end, part=part, delimiter=delimiter, previous=None,
                                         is_last=False, parts=parts)

    def _parse_array_multipaths_query_part(
            self, start: int, delimiter: str, max_end: int = 0) -> MultipathsArrayQueryPart:
        """Parse a multipaths object query part.

        Arguments:
            start: the index of the start of the path in the query.
            delimiter: the delimiter before the query part.
            max_end: an optional last position up to where the multipaths can extend.

        Returns:
            the multipaths array query part.

        Raises:
            gjson.GJSONParseError: on invalid query.

        """
        end = self._find_closing_parentheses(start=start, opening='[', max_end=max_end)
        part = self._query[start:end + 1]
        parts = []
        skip_until = 0

        for i, _ in enumerate(part[1:-1], start=1):
            if skip_until and i <= skip_until:
                if i == skip_until:
                    skip_until = 0

                continue

            try:
                values = self._parse(
                    start=start + i,
                    end=end - 1,
                    max_end=max_end - 1 if max_end else end - 1,
                    delimiter=delimiter,
                    in_multipaths=True)
            except GJSONInvalidSyntaxError:
                raise
            except GJSONParseError:  # In multipaths, paths that fails are silently suppressed
                values = []

            if values:
                parts.append(values)
                skip_until = values[-1].end - start + 1
            else:
                skip_until = end - start

        return MultipathsArrayQueryPart(start=start, end=end, part=part, delimiter=delimiter, previous=None,
                                        is_last=False, parts=parts)

    def _parse_literal_query_part(self, start: int, delimiter: str, max_end: int = 0,
                                  in_multipaths: bool = False) -> LiteralQueryPart:
        """Parse a literal query part.

        Arguments:
            start: the index of the start of the path in the query.
            delimiter: the delimiter before the query part.
            max_end: an optional last position up to where the multipaths can extend.
            in_multipaths: whether the part to be parsed is inside a multipaths.

        Returns:
            the literal query part.

        Raises:
            gjson.GJSONParseError: on invalid query.

        """
        end = -1
        begin = self._query[start + 1:start + 2]
        if begin in ('{', '['):
            end = self._find_closing_parentheses(start=start + 1, opening=begin, max_end=max_end)

        elif begin == '"':
            query = self._query[start + 2:max_end + 1] if max_end else self._query[start + 2:]
            match = re.search(r'(?<!\\)(")', query)  # Negative lookbehind assertion
            if match is None or match.end() == -1:
                raise GJSONParseError('Unable to find end of literal string.',
                                      query=self._query, position=start + 2)

            # match.end() returns the next position after the last matched character
            end = start + 1 + match.end()

        else:
            query = self._query[start + 1:max_end + 1] if max_end else self._query[start + 1:]
            if match := re.match(r'(true|false|null|NaN|(-)?Infinity)', query):
                # Includes also Infinite and NaN values:
                # https://docs.python.org/3/library/json.html#infinite-and-nan-number-values
                offset = 0
            elif match := re.match(r'-?(0|[1-9][0-9]*)(.[0-9]+)?((e|E)(\+|-|)[0-9]+)?', query):
                offset = 0  # JSON number
            else:  # Catch until the first delimiter or end of string
                delimiters = MULTIPATHS_DELIMITERS if in_multipaths else DELIMITERS
                pattern = '|'.join(re.escape(delimiter) for delimiter in delimiters)
                match = re.search(fr'(?<!\\)({pattern}|$)', query)  # Negative lookbehind assertion
                offset = 1 if match and match.group() else 0

                if match is None or match.end() == -1:  # pragma: nocover the above regex always matches
                    raise GJSONParseError('Invalid JSON literal.', query=self._query, position=start + 1)

            end = start + match.end() - offset  # match.end() gives the next position after the last matched character

        # Allow for empty literals '!' at this point to be able to handle them differently while parsing the parts
        # based if it's in a multipaths or after a query.
        part = self._query[start:end + 1]
        return LiteralQueryPart(start=start, end=end, part=part, delimiter=delimiter, previous=None, is_last=False)

    def _parse_part(self, part: BaseQueryPart, obj: Any, in_multipaths: bool = False) -> Any:
        """Parse the given part of the full query.

        Arguments:
            part: the query part as already parsed.
            obj: the current object.
            in_multipaths: whether the part to be parsed is inside a multipaths.

        Raises:
            gjson.GJSONParseError: on invalid query.

        Returns:
            the result of the query.

        """
        in_hash = False
        in_query_all = False
        ret: Any

        if isinstance(obj, NoResult):
            return obj

        if isinstance(part, ArrayLenghtQueryPart):
            in_hash = True
            if part.is_last:
                if part.delimiter == DOT_DELIMITER and (self._after_hash or self._after_query_all):
                    ret = []
                elif part.delimiter == PIPE_DELIMITER and isinstance(part.previous, ArrayLenghtQueryPart):
                    raise GJSONParseError('The pipe delimiter cannot immediately follow the # element.',
                                          query=self._query, position=part.start)
                elif self._is_sequence(obj):
                    ret = len(obj)
                else:
                    raise GJSONParseError('Expected a sequence like object for query part # at the end of the query, '
                                          f'got {type(obj)}.', query=self._query, position=part.start)
            else:
                ret = obj

        elif isinstance(part, ArrayQueryQueryPart):
            if not self._is_sequence(obj):
                raise GJSONParseError(f'Queries are supported only for sequence like objects, got {type(obj)}.',
                                      query=self._query, position=part.start)

            in_query_all = not part.first_only
            ret = self._parse_query(part, obj)

        elif isinstance(part, ModifierQueryPart):
            ret = self._apply_modifier(part, obj)

        elif isinstance(part, ArrayIndexQueryPart):
            if isinstance(obj, Mapping):  # Integer object keys not supported by JSON
                if not in_multipaths and part.part not in obj:
                    raise GJSONParseError(f'Mapping object does not have key `{part}`.',
                                          query=self._query, position=part.start)
                ret = obj.get(part.part, NoResult())
            elif self._is_sequence(obj):
                if (self._after_hash or self._after_query_all) and part.delimiter == DOT_DELIMITER:
                    # Skip non mapping items and items without the given key
                    ret = [i[part.part] for i in obj if isinstance(i, Mapping) and part.part in i]
                elif (self._after_hash and part.delimiter == PIPE_DELIMITER
                        and isinstance(part.previous, ArrayLenghtQueryPart)):
                    raise GJSONParseError('Integer query part after a pipe delimiter on an sequence like object.',
                                          query=self._query, position=part.start)
                else:
                    num = len(obj)
                    if part.index >= num:
                        raise GJSONParseError(f'Index `{part}` out of range for sequence object with {num} items in '
                                              'query.', query=self._query, position=part.start)
                    ret = obj[part.index]
            else:
                raise GJSONParseError(f'Integer query part on unsupported object type {type(obj)}, expected a mapping '
                                      'or sequence like object.', query=self._query, position=part.start)

        elif isinstance(part, FieldQueryPart):
            if re.search(r'(?<!\\)(\?|\*)', part.part):  # Wildcards
                if not isinstance(obj, Mapping):
                    raise GJSONParseError(f'Wildcard matching key `{part}` requires a mapping object, got {type(obj)} '
                                          'instead.', query=self._query, position=part.start)

                input_parts = re.split(r'(?<!\\)(\?|\*)', part.part)
                pattern_parts = ['^']
                for input_part in input_parts:
                    if not input_part:
                        continue
                    if input_part == '*':
                        if pattern_parts[-1] != '.*':  # Squash all consecutive * to avoid re.match() performance issue
                            pattern_parts.append('.*')
                    elif input_part == '?':
                        pattern_parts.append('.')
                    else:
                        pattern_parts.append(re.escape(re.sub(r'\\(\*|\?)', r'\1', input_part)))
                pattern_parts.append('$')
                pattern = ''.join(pattern_parts)

                for key in obj.keys():
                    if re.match(pattern, key):
                        ret = obj[key]
                        break
                else:
                    raise GJSONParseError(f'No key matching pattern with wildcard `{part}`.',
                                          query=self._query, position=part.start)

            else:
                key = part.part.replace(ESCAPE_CHARACTER, '')
                failed = False
                if not self._after_hash and isinstance(obj, Mapping):
                    if not in_multipaths and key not in obj:
                        raise GJSONParseError(f'Mapping object does not have key `{key}`.',
                                              query=self._query, position=part.start)
                    ret = obj.get(key, NoResult())
                elif (self._after_hash or self._after_query_all) and part.delimiter == DOT_DELIMITER:
                    if self._is_sequence(obj):
                        # Skip non mapping items and items without the given key
                        ret = [i[key] for i in obj if isinstance(i, Mapping) and key in i]
                    elif in_multipaths and isinstance(obj, Mapping):
                        ret = obj.get(key, NoResult())
                    elif in_multipaths:
                        ret = NoResult()
                    else:
                        failed = True
                else:
                    failed = True

                if failed:
                    raise GJSONParseError(f'Invalid or unsupported query part `{key}`.',
                                          query=self._query, position=part.start)

        elif isinstance(part, MultipathsObjectQueryPart):
            if ((self._after_hash or self._after_query_all) and part.delimiter == DOT_DELIMITER
                    and self._is_sequence(obj)):
                ret = []
                for i in obj:
                    obj_item = {}
                    for obj_part in part.parts:
                        obj_ret = i
                        for obj_value in obj_part.values:
                            obj_ret = self._parse_part(obj_value, obj_ret, in_multipaths=True)

                        if not isinstance(obj_ret, NoResult):
                            obj_item[obj_part.key] = obj_ret

                    ret.append(obj_item)
            else:
                ret = {}
                for obj_part in part.parts:
                    obj_ret = obj
                    for obj_value in obj_part.values:
                        obj_ret = self._parse_part(obj_value, obj_ret, in_multipaths=True)

                    if not isinstance(obj_ret, NoResult):
                        ret[obj_part.key] = obj_ret

        elif isinstance(part, MultipathsArrayQueryPart):
            if ((self._after_hash or self._after_query_all) and part.delimiter == DOT_DELIMITER
                    and self._is_sequence(obj)):
                ret = []
                for i in obj:
                    array_item = []
                    for array_part in part.parts:
                        array_ret = i
                        for array_value in array_part:
                            array_ret = self._parse_part(array_value, array_ret, in_multipaths=True)

                        if not isinstance(array_ret, NoResult):
                            array_item.append(array_ret)

                    ret.append(array_item)
            else:
                ret = []
                for array_part in part.parts:
                    array_ret = obj
                    for array_value in array_part:
                        array_ret = self._parse_part(array_value, array_ret, in_multipaths=True)

                    if not isinstance(array_ret, NoResult):
                        ret.append(array_ret)

        elif isinstance(part, LiteralQueryPart):
            try:
                new_obj = json.loads(part.part[1:], strict=False)
                json_error = ''
            except json.JSONDecodeError as ex:
                json_error = str(ex)
                new_obj = NoResult()

            ret = new_obj
            if (self._after_hash or self._after_query_all) and self._is_sequence(obj):
                if part.delimiter == DOT_DELIMITER:
                    if isinstance(new_obj, NoResult):
                        ret = []
                    else:
                        ret = [new_obj for _ in obj]
                elif part.delimiter == PIPE_DELIMITER:
                    ret = new_obj
            else:
                if part.delimiter == DOT_DELIMITER and not isinstance(new_obj, NoResult):
                    json_error = 'literal afer a dot delimiter.'
                    ret = NoResult()
                elif part.delimiter == PIPE_DELIMITER:
                    ret = new_obj

            if not in_multipaths and isinstance(ret, NoResult):
                raise GJSONParseError(
                    f'Unable to load literal JSON: {json_error}', query=self._query, position=part.start)

        if in_hash:
            self._after_hash = True
        if in_query_all:
            self._after_query_all = True

        return ret

    def _evaluate_query_return_value(self, query: ArrayQueryQueryPart, obj: Any) -> Any:
        """Evaluate the return value of an inline query #(...) / #(...)# depending on first match or all matches.

        Arguments:
            query: the query part.
            obj: the current object.

        Raises:
            gjson.GJSONParseError: if the query is for the first element and there are no matching items.

        Returns:
            the result of the query.

        """
        if query.first_only:
            if obj:
                return obj[0]

            raise GJSONParseError('Query for first element does not match anything.',
                                  query=self._query, position=query.start)

        return obj

    def _parse_query(self, query: ArrayQueryQueryPart, obj: Any) -> Any:
        """Parse an inline query #(...) / #(...)#.

        Arguments:
            query: the query part.
            obj: the current object.

        Raises:
            gjson.GJSONParseError: on invalid query.

        Returns:
            the result of the query.

        """
        if isinstance(query.value, ArrayQueryQueryPart):
            ret = []
            for i in obj:
                nested_obj = None
                if query.field:
                    if isinstance(i, Mapping) and query.field in i:
                        nested_obj = i[query.field]
                elif self._is_sequence(i):
                    nested_obj = i

                if nested_obj is not None and self._parse_query(query.value, nested_obj):
                    ret.append(i)

            return self._evaluate_query_return_value(query, ret)

        if not query.operator:
            return self._evaluate_query_return_value(query, [i for i in obj if query.field in i])

        key = query.field.replace('\\', '')
        try:
            value = json.loads(query.value, strict=False)
        except json.JSONDecodeError as ex:
            position = query.start + len(query.field) + len(query.operator)
            raise GJSONParseError(f'Invalid value `{query.value}` for the query key `{key}`.',
                                  query=self._query, position=position) from ex

        if not key and query.first_only and obj and isinstance(obj[0], Mapping):
            raise GJSONParseError('Query on mapping like objects require a key before the operator.',
                                  query=self._query, position=query.start)

        oper: Callable[[Any, Any], bool]
        if query.operator == '==~':
            if value not in (True, False):
                if query.first_only:
                    raise GJSONParseError(f'Queries ==~ operator requires a boolean value, got {type(value)} instead: '
                                          f'`{value}`.', query=self._query, position=query.start + len(query.field))

                return []

            def truthy_op(obj_a: Any, obj_b: bool) -> bool:
                truthy = operator.truth(obj_a)
                if obj_b:
                    return truthy
                return not truthy

            oper = truthy_op
        elif query.operator in ('==', '='):
            oper = operator.eq
        elif query.operator == '!=':
            oper = operator.ne
        elif query.operator == '<':
            oper = operator.lt
        elif query.operator == '<=':
            oper = operator.le
        elif query.operator == '>':
            oper = operator.gt
        elif query.operator == '>=':
            oper = operator.ge
        elif query.operator in ('%', '!%'):
            value = str(value).replace('*', '.*').replace('?', '.')
            value = f'^{value}$'
            if query.operator == '%':
                def match_op(obj_a: Any, obj_b: Any) -> bool:
                    if not isinstance(obj_a, str):
                        return False
                    return re.match(obj_b, obj_a) is not None
                oper = match_op
            else:
                def not_match_op(obj_a: Any, obj_b: Any) -> bool:
                    if not isinstance(obj_a, str):
                        return False
                    return re.match(obj_b, obj_a) is None
                oper = not_match_op

        try:
            if key:
                if query.operator == '==~':  # Consider missing keys as falsy according to GJSON docs.
                    ret = [i for i in obj if oper(i.get(key), value)]
                else:
                    ret = [i for i in obj if key in i and oper(i[key], value)]
            else:  # Query on an array of non-objects, match them directly
                ret = [i for i in obj if oper(i, value)]
        except TypeError:
            ret = []

        return self._evaluate_query_return_value(query, ret)

    def _apply_modifier(self, modifier: ModifierQueryPart, obj: Any) -> Any:
        """Apply a modifier.

        Arguments:
            part: the modifier query part to parse.
            obj: the current object before applying the modifier.

        Raises:
            gjson.GJSONError: when the modifier raises an exception.
            gjson.GJSONParseError: on unknown modifier.

        Returns:
            the object modifier according to the modifier.

        """
        try:
            modifier_func = getattr(self, f'_apply_modifier_{modifier.name}')
        except AttributeError:
            modifier_func = self._custom_modifiers.get(modifier.name)
            if modifier_func is None:
                raise GJSONParseError(f'Unknown modifier @{modifier.name}.',
                                      query=self._query, position=modifier.start) from None

        try:
            return modifier_func(modifier.options, obj, last=modifier.is_last)
        except GJSONError:
            raise
        except Exception as ex:
            raise GJSONError(f'Modifier @{modifier.name} raised an exception.') from ex

    def _apply_modifier_reverse(self, _options: dict[str, Any], obj: Any, *, last: bool) -> Any:
        """Apply the @reverse modifier.

        Arguments:
            options: the eventual options for the modifier, currently unused.
            obj: the current object to reverse.
            last: whether this is the final part of the query.

        Returns:
            the reversed object. If the object cannot be reversed is returned untouched.

        """
        del last  # for pylint, unused argument
        if isinstance(obj, Mapping):
            return {k: obj[k] for k in reversed(obj.keys())}
        if self._is_sequence(obj):
            return obj[::-1]

        return obj

    def _apply_modifier_keys(self, _options: dict[str, Any], obj: Any, *, last: bool) -> Any:
        """Apply the @keys modifier.

        Arguments:
            options: the eventual options for the modifier, currently unused.
            obj: the current object to get the keys from.
            last: whether this is the final part of the query.

        Raises:
            gjson.GJSONError: if the current object does not have a keys() method.

        Returns:
            the current object keys as list.

        """
        del last  # for pylint, unused argument
        try:
            return list(obj.keys())
        except AttributeError as ex:
            raise GJSONError('The current object does not have a keys() method.') from ex

    def _apply_modifier_values(self, _options: dict[str, Any], obj: Any, *, last: bool) -> Any:
        """Apply the @values modifier.

        Arguments:
            options: the eventual options for the modifier, currently unused.
            obj: the current object to get the values from.
            last: whether this is the final part of the query.

        Raises:
            gjson.GJSONError: if the current object does not have a values() method.

        Returns:
            the current object values as list.

        """
        del last  # for pylint, unused argument
        try:
            return list(obj.values())
        except AttributeError as ex:
            raise GJSONError('The current object does not have a values() method.') from ex

    def _apply_modifier_ugly(self, _options: dict[str, Any], obj: Any, *, last: bool) -> Any:
        """Apply the @ugly modifier to condense the output.

        Arguments:
            options: the eventual options for the modifier, currently unused.
            obj: the current object to uglyfy.
            last: whether this is the final part of the query.

        Returns:
            the current object, unmodified.

        """
        del last  # for pylint, unused argument
        self._dump_params['separators'] = (',', ':')
        self._dump_params['indent'] = None
        return obj

    def _apply_modifier_pretty(self, options: dict[str, Any], obj: Any, *, last: bool) -> Any:
        """Apply the @pretty modifier to pretty-print the output.

        Arguments:
            options: the eventual options for the modifier.
            obj: the current object to prettyfy.
            last: whether this is the final part of the query.

        Returns:
            the current object, unmodified.

        """
        del last  # for pylint, unused argument
        self._dump_params['indent'] = options.get('indent', 2)
        self._dump_params['sort_keys'] = options.get('sortKeys', False)
        self._dump_params['prefix'] = options.get('prefix', '')
        return obj

    def _apply_modifier_ascii(self, _options: dict[str, Any], obj: Any, *, last: bool) -> Any:
        """Apply the @ascii modifier to have all non-ASCII characters escaped when dumping the object.

        Arguments:
            options: the eventual options for the modifier, currently unused.
            obj: the current object to sort.
            last: whether this is the final part of the query.

        Returns:
            the current object, unmodified.

        """
        del last  # for pylint, unused argument
        self._dump_params['ensure_ascii'] = True
        return obj

    def _apply_modifier_sort(self, _options: dict[str, Any], obj: Any, *, last: bool) -> Any:
        """Apply the @sort modifier, sorts mapping and sequences.

        Arguments:
            options: the eventual options for the modifier, currently unused.
            obj: the current object to sort.
            last: whether this is the final part of the query.

        Raises:
            gjson.GJSONError: if the current object is not sortable.

        Returns:
            the sorted object.

        """
        del last  # for pylint, unused argument
        if isinstance(obj, Mapping):
            return {k: obj[k] for k in sorted(obj.keys())}
        if self._is_sequence(obj):
            return sorted(obj)

        raise GJSONError(f'@sort modifier not supported for object of type {type(obj)}. '
                         'Expected a mapping or sequence like object.')

    def _apply_modifier_valid(self, _options: dict[str, Any], obj: Any, *, last: bool) -> Any:
        """Apply the @valid modifier, checking that the current object can be converted to JSON.

        Arguments:
            options: the eventual options for the modifier, currently unused.
            obj: the current element to validate.
            last: whether this is the final part of the query.

        Raises:
            gjson.GJSONError: if the current object cannot be converted to JSON.

        Returns:
            the current object, unmodified.

        """
        del last  # for pylint, unused argument
        try:
            json.dumps(obj, **self._dump_params)
        except Exception as ex:
            raise GJSONError('The current object cannot be converted to JSON.') from ex

        return obj

    def _apply_modifier_this(self, _options: dict[str, Any], obj: Any, *, last: bool) -> Any:
        """Apply the @this modifier, that returns the current object.

        Arguments:
            options: the eventual options for the modifier, currently unused.
            obj: the current element to return.
            last: whether this is the final part of the query.

        Returns:
            the current object, unmodified.

        """
        del last  # for pylint, unused argument
        return obj

    def _apply_modifier_fromstr(self, _options: dict[str, Any], obj: Any, *, last: bool) -> Any:
        """Apply the @fromstr modifier, converting a string to JSON, if valid.

        Arguments:
            options: the eventual options for the modifier, currently unused.
            obj: the current element from where to extract the JSON.
            last: whether this is the final part of the query.

        Raises:
            gjson.GJSONError: if the current object cannot be converted to JSON.

        Returns:
            the parsed JSON.

        """
        del last  # for pylint, unused argument
        if not isinstance(obj, (str, bytes)):
            raise GJSONError(f'Modifier @fromstr got object of type {type(obj)} as input, expected string or bytes.')

        try:
            return json.loads(obj, strict=False)
        except Exception as ex:
            raise GJSONError('The current @fromstr input object cannot be converted to JSON.') from ex

    def _apply_modifier_tostr(self, _options: dict[str, Any], obj: Any, *, last: bool) -> Any:
        """Apply the @tostr modifier, converting the current object to a JSON-encoded string, if valid.

        Arguments:
            options: the eventual options for the modifier, currently unused.
            obj: the current element from where to extract the JSON.
            last: whether this is the final part of the query.

        Raises:
            gjson.GJSONError: if the current object cannot be converted to a JSON-encoded string.

        Returns:
            the JSON-encoded string.

        """
        del last  # for pylint, unused argument
        try:
            return json.dumps(obj, ensure_ascii=False)
        except Exception as ex:
            raise GJSONError('The current object cannot be converted to a JSON-encoded string for @tostr.') from ex

    def _apply_modifier_group(self, _options: dict[str, Any], obj: Any, *, last: bool) -> Any:
        """Apply the @group modifier, that groups a dictionary of lists in a list of dictionaries.

        Example input::

            {"id": ["123", "456", "789"], "val": [2, 1]}

        Example output::

            [{"id": "123", "val": 2}, {"id": "456", "val": 1}, {"id": "789"}]

        Arguments:
            options: the eventual options for the modifier, currently unused.
            obj: the current element to group.
            last: whether this is the final part of the query.

        Raises:
            gjson.GJSONError: if the current object is not a dictionary.

        Returns:
            a list with the grouped objects or an empty list if the input has no lists as values.

        """
        del last  # for pylint, unused argument
        if not isinstance(obj, Mapping):
            raise GJSONError(f'Modifier @group got object of type {type(obj)} as input, expected dictionary.')

        # Skip all values that aren't lists:
        obj = {k: v for k, v in obj.items() if self._is_sequence(v)}
        # Fill missing values with NoResult to remove them afterwards
        obj = [dict(zip_longest(obj.keys(), values)) for values in zip_longest(*obj.values(), fillvalue=NoResult())]
        # Skip keys with value NoResult in each dictionary
        return [{k: v for k, v in i.items() if not isinstance(v, NoResult)} for i in obj]

    def _apply_modifier_join(self, _options: dict[str, Any], obj: Any, *, last: bool) -> Any:
        """Apply the @join modifier, that joins a list of dictionaries into a single dictionary.

        Items in the sequence that are not dictionaries are skipped.
        Differently from GJSON there is no support for duplicated keys as the can't exist in Python dictionaries.
        Hence this modifier doesn't accept any option.

        Example input::

            [{"first": "Tom", "age": 37}, {"age": 41}]

        Example output::

            {"first": "Tom", "age":41}

        Arguments:
            options: the eventual options for the modifier, currently unused.
            obj: the current element to join.
            last: whether this is the final part of the query.

        Returns:
            the object untouched if the object is not a sequence, a dictionary with joined objects otherwise.

        """
        del last  # for pylint, unused argument
        if not self._is_sequence(obj):
            return obj

        ret: dict[Any, Any] = {}
        for item in obj:
            if isinstance(item, Mapping):
                ret.update(item)

        return ret

    def _apply_modifier_top_n(self, options: dict[str, Any], obj: Any, *, last: bool) -> Any:
        """Apply the @top_n modifier to find the most common values of a given field.

        Arguments:
            options: the eventual modifier options. If not specified all items are returned. If specified it must
                contain a 'n' key with the number of top N to return.
            obj: the current object to extract the N most common items.
            last: whether this is the final part of the query.

        Raises:
            gjson.GJSONError: if the current object is not a sequence.

        Returns:
            dict: a dictionary of unique items as keys and the count as value.

        """
        del last  # for pylint, unused argument
        if not self._is_sequence(obj):
            raise GJSONError(f'@top_n modifier not supported for object of type {type(obj)}. '
                             'Expected a sequence like object.')

        return dict(Counter(obj).most_common(options.get('n')))

    def _apply_modifier_sum_n(self, options: dict[str, Any], obj: Any, *, last: bool) -> Any:
        """Apply the @sum_n modifier that groups the values of a given key while summing the values of another key.

        The key used to sum must have numeric values.

        Arguments:
            options: the modifier options. It must contain a 'group' key with the name of the field to use to group the
                items as value and a 'sum' key with the name of the field to use to sum the values for each unique
                grouped identifier. If a 'n' key is also provided, only the top N results are returned. If not
                specified all items are returned.
            obj: the current object to group and sum the top N values.
            last: whether this is the final part of the query.

        Raises:
            gjson.GJSONError: if the current object is not a sequence.

        Returns:
            dict: a dictionary of unique items as keys and the sum as value.

        """
        del last  # for pylint, unused argument
        if not self._is_sequence(obj):
            raise GJSONError(f'@sum_n modifier not supported for object of type {type(obj)}. '
                             'Expected a sequence like object.')

        results: Counter[Any] = Counter()
        for item in obj:
            results[item[options['group']]] += item[options['sum']]
        return dict(results.most_common(options.get('n')))

    def _apply_modifier_flatten(self, options: dict[str, Any], obj: Any, *, last: bool) -> Any:
        """Apply the @flatten modifier.

        Arguments:
            options: the eventual modifier options.
            obj: the current object to flatten.
            last: whether this is the final part of the query.

        Returns:
            the modified object.

        """
        del last  # for pylint, unused argument
        if not self._is_sequence(obj):
            return obj

        return list(self._flatten_sequence(obj, deep=options.get('deep', False)))

    def _flatten_sequence(self, obj: Any, deep: bool = False) -> Any:
        """Flatten nested sequences in the given object.

        Arguments:
            obj: the current object to flatten
            deep: if :py:data:`True` recursively flatten nested sequences. By default only the first level is
                processed.

        Returns:
            the flattened object if it was a flattable sequence, the given object itself otherwise.

        """
        for elem in obj:
            if self._is_sequence(elem):
                if deep:
                    yield from self._flatten_sequence(elem, deep=deep)
                else:
                    yield from elem
            else:
                yield elem
